from neuralogic.dataset import Data, TensorDataset, Dataset, Sample
from neuralogic.core import Template, Settings, Relation
from neuralogic.nn.module import GCNConv
from neuralogic.optim import SGD


def test_quick_start_from_tensor():
    data = Data(
        edge_index=[
            [0, 1, 1, 2, 2, 0],
            [1, 0, 2, 1, 0, 2],
        ],
        x=[[0], [1], [-1]],
        y=[[1], [0], [1]],
        y_mask=[0, 1, 2],
    )

    dataset = TensorDataset(data=[data])

    assert len(dataset.data) == 1

    logic_dataset = dataset.to_dataset()

    print("--")
    print(logic_dataset)
    for x in logic_dataset.samples:
        print([str(e) for e in x.example])
        print(x.query)

    assert len(logic_dataset) == 3
    assert len(logic_dataset[0]) == 9

    expected = [
        "<1> edge(0, 1).",
        "<1> edge(1, 0).",
        "<1> edge(1, 2).",
        "<1> edge(2, 1).",
        "<1> edge(2, 0).",
        "<1> edge(0, 2).",
        "<0> node_feature(0).",
        "<1> node_feature(1).",
        "<-1> node_feature(2).",
    ]

    for a, b in zip(logic_dataset[0].example, expected):
        assert str(a) == b

    assert str(logic_dataset[0]) == "1.0 predict(0)."
    assert str(logic_dataset[1]) == "0.0 predict(1)."
    assert str(logic_dataset[2]) == "1.0 predict(2)."


def test_model_evaluation_from_tensor():
    data = Data(
        edge_index=[
            [0, 1, 1, 2, 2, 0],
            [1, 0, 2, 1, 0, 2],
        ],
        x=[[0], [1], [-1]],
        y=[[1], [0], [1]],
        y_mask=[0, 1, 2],
    )

    dataset = TensorDataset(data=[data])

    template = Template()
    template.add_module(
        GCNConv(in_channels=1, out_channels=5, output_name="h0", feature_name="node_feature", edge_name="edge")
    )
    template.add_module(
        GCNConv(in_channels=5, out_channels=1, output_name="predict", feature_name="h0", edge_name="edge")
    )

    settings = Settings(optimizer=SGD(0.01))
    template.build(settings)
    built_dataset = template.build_dataset(dataset)

    # or template.test() to change the mode
    output = template.train(built_dataset, epochs=100)

    assert len(output[0]) == 3
    assert len(output[1]) == 3


def test_model_evaluation_from_logic():
    dataset = Dataset()

    example = [
        Relation.edge(0, 1),
        Relation.edge(1, 2),
        Relation.edge(2, 0),
        Relation.edge(1, 0),
        Relation.edge(2, 1),
        Relation.edge(0, 2),
        Relation.node_feature(0)[0],
        Relation.node_feature(1)[1],
        Relation.node_feature(2)[-1],
    ]

    dataset.add_samples(
        [
            Sample(Relation.predict(0)[1], example),
            Sample(Relation.predict(1)[0], example),
            Sample(Relation.predict(2)[1], example),
        ]
    )

    template = Template()
    template.add_module(
        GCNConv(in_channels=1, out_channels=5, output_name="h0", feature_name="node_feature", edge_name="edge")
    )
    template.add_module(
        GCNConv(in_channels=5, out_channels=1, output_name="predict", feature_name="h0", edge_name="edge")
    )

    settings = Settings(optimizer=SGD(0.01))
    template.build(settings)
    built_dataset = template.build_dataset(dataset)

    # or template.test() to change the mode
    output = template.train(built_dataset, epochs=100)

    assert len(output[0]) == 3
    assert len(output[1]) == 3



from neuralogic.dataset.base import BaseDataset, ConvertibleDataset
from neuralogic.dataset.file import FileDataset
from neuralogic.dataset.logic import Dataset, Sample
from neuralogic.dataset.tensor import TensorDataset, Data
from neuralogic.dataset.csv import Mode, CSVFile, CSVDataset
from neuralogic.dataset.db import DBDataset, DBSource

__all__ = [
    "BaseDataset",
    "ConvertibleDataset",
    "Sample",
    "FileDataset",
    "Dataset",
    "TensorDataset",
    "Data",
    "Mode",
    "CSVFile",
    "CSVDataset",
    "DBDataset",
    "DBSource",
]



from typing import List

import pytest

from neuralogic import initialize, is_initialized
from neuralogic.optim import Adam
from neuralogic.optim.lr_scheduler import GeometricLR, ArithmeticLR


@pytest.mark.parametrize(
    "lr, max_steps, expected",
    [
        (100, 100, [99, 98, 97, 96]),
        (100, 10, [90, 80, 70, 60]),
    ],
)
def test_arithmetic_lr_decay(lr: float, max_steps: int, expected: List[float]):
    if not is_initialized():
        initialize()

    optimizer = Adam(lr=lr, lr_decay=ArithmeticLR(max_steps))
    optimizer.initialize()

    assert optimizer.lr == lr

    lr_decay = optimizer._lr_decay

    for exp in expected:
        lr_decay.decay(1)
        assert optimizer.lr == exp


def test_geometric_lr_decay():
    pass


@pytest.mark.parametrize(
    "lr, steps, rate, expected",
    [
        (100, 1, 1, [100, 100, 100, 100]),
        (100, 1, 0.9, [90, 90 * 0.9, 90 * (0.9**2), 90 * (0.9**3)]),
        (100, 2, 0.9, [100, 90, 90, 90 * 0.9, 90 * 0.9, 90 * (0.9**2), 90 * (0.9**2)]),
    ],
)
def test_geometric_lr_decay(lr: float, steps: int, rate: float, expected: List[float]):
    if not is_initialized():
        initialize()

    optimizer = Adam(lr=lr, lr_decay=GeometricLR(rate, steps))
    optimizer.initialize()

    assert optimizer.lr == lr

    lr_decay = optimizer._lr_decay

    for i, exp in enumerate(expected):
        lr_decay.decay(i + 1)
        assert optimizer.lr == exp



from neuralogic.core import Template, R, V, C


def test_inference_engine_london_reachable() -> None:
    """
    Test of the inference engine
    based on https://book.simply-logical.space/part_i.html#a_brief_introduction_to_clausal_logic
    """
    template = Template()

    template.add_rules(
        [
            R.connected(C.bond_street, C.oxford_circus, C.central),
            R.connected(C.oxford_circus, C.tottenham_court_road, C.central),
            R.connected(C.bond_street, C.green_park, C.jubilee),
            R.connected(C.green_park, C.charing_cross, C.jubilee),
            R.connected(C.green_park, C.piccadilly_circus, C.piccadilly),
            R.connected(C.piccadilly_circus, C.leicester_square, C.piccadilly),
            R.connected(C.green_park, C.oxford_circus, C.victoria),
            R.connected(C.oxford_circus, C.piccadilly_circus, C.bakerloo),
            R.connected(C.piccadilly_circus, C.charing_cross, C.bakerloo),
            R.connected(C.tottenham_court_road, C.leicester_square, C.northern),
            R.connected(C.leicester_square, C.charing_cross, C.northern),
            R.reachable(V.X, V.Y) <= R.connected(V.X, V.Y, V.L),
            R.reachable(V.X, V.Y) <= (R.connected(V.X, V.Z, V.L), R.reachable(V.Z, V.Y)),
        ]
    )

    # ask if tottenham_court_road can be reached from green_park
    assert template.query(R.reachable(C.green_park, C.tottenham_court_road))

    # green_park cannot be reached from charing_cross
    # random_place does not exist in the dataset (cannot be reached from anywhere)
    assert not template.query(R.reachable(C.charing_cross, C.green_park))
    assert not template.q(R.reachable(C.charing_cross, C.random_place))


def test_inference_engine_london() -> None:
    """
    Test of the inference engine
    based on https://book.simply-logical.space/part_i.html#a_brief_introduction_to_clausal_logic
    """
    template = Template()

    knowledge = [
        R.connected(C.bond_street, C.oxford_circus, C.central),
        R.connected(C.oxford_circus, C.tottenham_court_road, C.central),
        R.connected(C.bond_street, C.green_park, C.jubilee),
        R.connected(C.green_park, C.charing_cross, C.jubilee),
        R.connected(C.green_park, C.piccadilly_circus, C.piccadilly),
        R.connected(C.piccadilly_circus, C.leicester_square, C.piccadilly),
        R.connected(C.green_park, C.oxford_circus, C.victoria),
        R.connected(C.oxford_circus, C.piccadilly_circus, C.bakerloo),
        R.connected(C.piccadilly_circus, C.charing_cross, C.bakerloo),
        R.connected(C.tottenham_court_road, C.leicester_square, C.northern),
        R.connected(C.leicester_square, C.charing_cross, C.northern),
    ]

    template.add_rules(
        [
            R.nearby(V.X, V.Y) <= R.connected(V.X, V.Y, V.L),
            R.nearby(V.X, V.Y) <= (R.connected(V.X, V.Z, V.L), R.connected(V.Z, V.Y, V.L)),
        ]
    )

    engine = template.build()

    # Run query for nearby(X, oxford_circus)
    # Should yield two substitutions for x (green_park and bond_street)
    substitutions = list(engine.q(R.nearby(V.X, C.oxford_circus), knowledge))

    assert substitutions[0]["X"] == "bond_street"
    assert substitutions[1]["X"] == "green_park"
    assert len(substitutions) == 2 and len(substitutions[0]) == 1 and len(substitutions[1]) == 1

    # Run query for nearby(X, tottenham_court_road)
    substitutions = sorted(engine.q(R.nearby(V.X, C.charing_cross), knowledge), key=lambda x: x["X"])

    assert substitutions[0]["X"] == "bond_street"
    assert substitutions[1]["X"] == "green_park"
    assert substitutions[2]["X"] == "leicester_square"
    assert substitutions[3]["X"] == "oxford_circus"
    assert substitutions[4]["X"] == "piccadilly_circus"
    assert substitutions[5]["X"] == "tottenham_court_road"
    assert len(substitutions) == 6 and len(substitutions[0]) == 1 and len(substitutions[1]) == 1

    # Run query for connected(X, leicester_square, Z)
    # Should yield two substitutions:
    # {'X': 'piccadilly_circus', 'Z': 'piccadilly'}, {'X': 'tottenham_court_road', 'Z': 'northern'}
    substitutions = sorted(list(engine.q(R.connected(V.X, C.leicester_square, V.Z), knowledge)), key=lambda a: a["X"])

    assert substitutions[0]["X"] == "piccadilly_circus"
    assert substitutions[0]["Z"] == "piccadilly"

    assert substitutions[1]["X"] == "tottenham_court_road"
    assert substitutions[1]["Z"] == "northern"
    assert len(substitutions) == 2 and len(substitutions[0]) == 2 and len(substitutions[1]) == 2


def test_listing_all_queries() -> None:
    template = Template()

    template += R.h(V.X) <= R.edge(V.Y, V.X)
    template += R.h1(V.X) <= (R.h(V.Y), R.edge(V.Y, V.X))
    template += R.q <= R.h1(V.X)

    template += R.edge(1, 2)
    template += R.edge(2, 3)
    template += R.edge(3, 1)

    queries = list(template.derivable_queries())

    expected_queries = sorted(
        ["edge(1, 2).", "edge(2, 3).", "edge(3, 1).", "h(2).", "h(3).", "h(1).", "h1(1).", "h1(2).", "h1(3).", "q."]
    )
    str_queries = sorted([str(query) for query in queries])

    for a, b in zip(expected_queries, str_queries):
        assert a == b

    queries = list(template.derivable_queries([R.edge(1, 4)]))

    expected_queries = sorted(
        [
            "edge(1, 2).",
            "edge(2, 3).",
            "edge(3, 1).",
            "edge(1, 4).",
            "h(2).",
            "h(3).",
            "h(1).",
            "h(4).",
            "h1(1).",
            "h1(2).",
            "h1(3).",
            "h1(4).",
            "q.",
        ]
    )
    str_queries = sorted([str(query) for query in queries])

    print(str_queries)
    print(expected_queries)

    for a, b in zip(expected_queries, str_queries):
        assert a == b



import dataclasses
import os
import sys
from enum import Enum
from typing import List, Tuple, Any

import jpype


_is_logging_initialized = False
_default_logging = True
_loggers_buffer: List[Tuple] = []


class TextIOWrapper:
    def __init__(self, wrapped_text_io):
        self.wrapped_text_io = wrapped_text_io

    def write(self, string):
        self.wrapped_text_io.write(str(string))


class Level(Enum):
    """Logging level"""

    OFF = 2147483647
    SEVERE = 1000
    WARNING = 900
    INFO = 800
    CONFIG = 700
    FINE = 500
    FINER = 400
    FINEST = 300
    ALL = -2147483648


class Formatter(Enum):
    """Logged information formatters"""

    COLOR = "color"
    NORMAL = "normal"


@dataclasses.dataclass
class LogHandler:
    output: Any
    level: Level = Level.FINER
    formatter: Formatter = Formatter.COLOR


def _init_logging():
    global _is_logging_initialized

    jpype.java.lang.System.setOut(jpype.java.io.PrintStream(os.devnull))
    jpype.java.lang.System.setErr(jpype.java.io.PrintStream(os.devnull))

    settings_class = jpype.JClass("cz.cvut.fel.ida.setup.Settings")
    settings = settings_class()
    settings.supressConsoleOutput = True
    settings.supressLogFileOutput = True
    settings.loggingLevel = jpype.JClass("java.util.logging.Level").OFF

    jpype.JClass("cz.cvut.fel.ida.logging.Logging").initLogging(settings)
    _is_logging_initialized = True

    if _default_logging:
        add_handler(sys.stdout, Level.INFO)

    for handler_settings in _loggers_buffer:
        add_handler(*handler_settings)
    _loggers_buffer.clear()


def add_log_handler(handler: LogHandler):
    """
    Add logger handler for an insight into the java backend. Overrides the default logger to stdout.

    :param handler: LogHandler
    """
    return add_handler(handler.output, handler.level, handler.formatter)


def add_handler(output, level: Level = Level.FINER, formatter: Formatter = Formatter.COLOR):
    """
    Add logger handler for an insight into the java backend. Overrides the default logger to stdout.

    :param output: File-like object (has ``write(text: str)`` method)
    :param level: The logging level
    :param formatter: The log formatter
    """
    global _default_logging
    _default_logging = False

    if not _is_logging_initialized:
        _loggers_buffer.append((output, level, formatter))
        return

    root_logger = jpype.java.util.logging.Logger.getLogger("")
    wrapped = TextIOWrapper(output)

    java_io_wrapper = jpype.JProxy("cz.cvut.fel.ida.utils.python.PythonOutputStream.TextIOWrapper", inst=wrapped)
    java_output_stream = jpype.JClass("cz.cvut.fel.ida.utils.python.PythonOutputStream")(java_io_wrapper)

    if formatter == Formatter.COLOR:
        java_formatter = jpype.JClass("cz.cvut.fel.ida.logging.ColoredFormatter")()
    elif formatter == Formatter.NORMAL:
        java_formatter = jpype.JClass("cz.cvut.fel.ida.logging.NormalFormatter")()
    else:
        raise NotImplementedError(f"Unknown formatter {formatter}")

    print_stream = jpype.java.io.PrintStream(java_output_stream)
    stream_handler = jpype.JClass("cz.cvut.fel.ida.logging.FlushStreamHandler")(print_stream, java_formatter)
    stream_handler.setLevel(getattr(jpype.java.util.logging.Level, str(level.name)))

    if level.value < root_logger.getLevel().intValue():
        root_logger.setLevel(getattr(jpype.java.util.logging.Level, str(level.name)))

    root_logger.addHandler(stream_handler)


def clear_handlers():
    """Clear all handlers"""
    global _default_logging
    _default_logging = False

    if not _is_logging_initialized:
        _loggers_buffer.clear()
        return

    root_logger = jpype.java.util.logging.Logger.getLogger("")

    for handler in root_logger.getHandlers():
        root_logger.removeHandler(handler)


__all__ = ["add_handler", "add_log_handler", "clear_handlers", "Level", "Formatter", "TextIOWrapper", "LogHandler"]



from typing import Sequence

import jpype

from neuralogic.core.constructs.function.function import AggregationFunction


class SoftmaxAggregation(AggregationFunction):
    __slots__ = ("agg_terms", "var_terms")

    def __init__(
        self,
        name: str,
        *,
        agg_terms: Sequence[str] | None = None,
    ):
        super().__init__(name)
        self.term_indices = agg_terms
        self.agg_terms = agg_terms

    def __call__(self, *, agg_terms: Sequence[int] | None = None):
        softmax = SoftmaxAggregation(self.name, agg_terms=agg_terms)
        return AggregationFunction.__call__(softmax)

    def is_parametrized(self) -> bool:
        return self.agg_terms is not None

    def get(self):
        return jpype.JClass("cz.cvut.fel.ida.algebra.functions.combination.Softmax")(self.term_indices)

    def __str__(self):
        if self.agg_terms is None:
            return "softmax"
        return f"softmax(agg_terms=[{', '.join(self.agg_terms)}])"

    def rule_head_dependant(self) -> bool:
        return self.agg_terms is not None

    def process_head(self, head) -> "SoftmaxAggregation":
        term_indices = []

        for agg_term in set(self.agg_terms):
            if not agg_term[0].isupper():
                raise NotImplementedError(f"Softmax aggregable terms can be only variables. Provided: {agg_term}")

            for i, term in enumerate(head.terms):
                if agg_term == term:
                    term_indices.append(i)
                    break

        aggregation = SoftmaxAggregation(self.name, agg_terms=self.agg_terms)
        aggregation.term_indices = term_indices

        return aggregation



import math

import numpy as np

from neuralogic.core.constructs.function import Combination
from neuralogic.core.constructs.factories import R
from neuralogic.nn.module.module import Module


class PositionalEncoding(Module):
    def __init__(
        self,
        embed_dim: int,
        max_len: int,
        output_name: str,
        input_name: str,
        arity: int = 1,
        learnable: bool = False,
    ):
        self.embed_dim = embed_dim
        self.max_len = max_len

        self.output_name = output_name
        self.input_name = input_name

        self.arity = arity
        self.learnable = learnable

    def __call__(self):
        terms = [f"X{i}" for i in range(self.arity - 1)]
        all_terms = [f"X{i}" for i in range(self.arity)]

        position = np.arange(0, self.max_len).reshape((self.max_len, 1))
        div_term = np.exp(np.arange(0, self.embed_dim, 2) * (-math.log(1000.0) / self.embed_dim))

        pe = np.zeros((self.max_len, self.embed_dim))
        mul = position * div_term

        pe[:, 0::2] = np.sin(mul)
        pe[:, 1::2] = np.cos(mul)

        pe_rel = R.get(f"{self.output_name}__pe")
        out_rel = R.get(self.output_name)
        in_rel = R.get(self.input_name)

        if self.learnable:
            rules = [pe_rel(*terms, i)[row] for i, row in enumerate(pe)]
        else:
            rules = [pe_rel(*terms, i)[row].fixed() for i, row in enumerate(pe)]

        rules.append((out_rel(all_terms) <= (pe_rel(all_terms), in_rel(all_terms))) | [Combination.SUM])

        return rules



from neuralogic.core.neural_module import NeuralModule


class Trainer:
    def __init__(self, module: NeuralModule):
        self._module = module

    def train(self, dataset, epochs):
        return self._module.train(dataset, epochs)

    def test(self, dataset):
        return self._module.test(dataset)



import numpy as np
import pytest
import torch
from torch_geometric.data import Data
from torch_geometric.nn import GENConv, GINEConv, GCNConv

from neuralogic.core import Template, Settings, R, Aggregation
from neuralogic.dataset import Dataset, Sample
from neuralogic.nn.loss import MSE

import neuralogic.nn.module
from neuralogic.optim import Adam


@pytest.mark.parametrize(
    "input_size, hidden_size",
    [
        (5, 1),
        (10, 1),
    ],
)
def test_gen_module(input_size, hidden_size):
    """Test that PyNeuraLogic GENConv layer computes the same as PyG GENConv layer"""

    edge_index = torch.tensor([[0, 1], [1, 0], [1, 2], [2, 1]], dtype=torch.long)
    x = torch.randn(
        (
            3,
            input_size,
        ),
        dtype=torch.float,
    )
    e = torch.randn(
        (
            4,
            input_size,
        ),
        dtype=torch.float,
    )

    data = Data(x=x, edge_index=edge_index.t().contiguous(), edge_attr=e)

    gen = GENConv(input_size, hidden_size, aggr="mean", num_layers=1, eps=0, edge_dim=input_size)
    for m in gen.mlp._modules.values():
        m.bias = None

    template = Template()
    template += neuralogic.nn.module.GENConv(
        input_size, hidden_size, "h", "f", "e", num_layers=1, aggregation=Aggregation.AVG, eps=0, edge_dim=input_size
    )

    model = template.build(
        Settings(chain_pruning=False, iso_value_compression=False, optimizer=Adam(lr=0.001), error_function=MSE())
    )

    parameters = model.parameters()
    torch_parameters = [parameter.tolist() for parameter in gen.parameters()]

    parameters["weights"][2] = [torch_parameters[0][i] for i in range(0, hidden_size)]
    parameters["weights"][1] = [torch_parameters[1][i] for i in range(0, hidden_size)]
    parameters["weights"][3] = [torch_parameters[2][i] for i in range(0, hidden_size)]
    parameters["weights"][4] = torch_parameters[3][0]

    model.load_state_dict(parameters)

    example = [
        R.f(0)[x[0]],
        R.f(1)[x[1]],
        R.f(2)[x[2]],
        R.e(0, 1)[e[0]],
        R.e(1, 0)[e[1]],
        R.e(1, 2)[e[2]],
        R.e(2, 1)[e[3]],
    ]

    dataset = Dataset([Sample(R.h(i), example) for i in range(3)])
    bd = model.build_dataset(dataset)

    output = gen(x=data.x, edge_index=data.edge_index, edge_attr=data.edge_attr)
    result = model.test(bd)

    assert np.allclose([float(x) for x in output], [float(x) for x in result], atol=10e-5)


@pytest.mark.parametrize(
    "input_size",
    [
        5,
        10,
    ],
)
def test_gine_module(input_size):
    """Test that PyNeuraLogic GINEConv layer computes the same as PyG GINEConv layer"""

    edge_index = torch.tensor([[0, 1], [1, 0], [1, 2], [2, 1]], dtype=torch.long)
    x = torch.randn(
        (
            3,
            input_size,
        ),
        dtype=torch.float,
    )
    e = torch.randn(
        (
            4,
            input_size,
        ),
        dtype=torch.float,
    )

    data = Data(x=x, edge_index=edge_index.t().contiguous(), edge_attr=e)

    class Identity(torch.nn.Module):
        def forward(self, x):
            return x

    gin = GINEConv(Identity())

    template = Template()
    template += neuralogic.nn.module.GINEConv(input_size, "f", "e", "h")

    model = template.build(
        Settings(chain_pruning=False, iso_value_compression=False, optimizer=Adam(lr=0.001), error_function=MSE())
    )

    example = [
        R.f(0)[x[0]],
        R.f(1)[x[1]],
        R.f(2)[x[2]],
        R.e(0, 1)[e[0]],
        R.e(1, 0)[e[1]],
        R.e(1, 2)[e[2]],
        R.e(2, 1)[e[3]],
    ]

    dataset = Dataset([Sample(R.h(i), example) for i in range(3)])
    bd = model.build_dataset(dataset)

    output = gin(x=data.x, edge_index=data.edge_index, edge_attr=data.edge_attr)
    result = model.test(bd)

    assert np.allclose([[float(x) for x in xs] for xs in output], [[float(x) for x in xs] for xs in result], atol=10e-5)


@pytest.mark.parametrize(
    "input_size, output_size",
    [
        (5, 2),
        (10, 2),
    ],
)
def test_gcn_module(input_size, output_size):
    """Test that PyNeuraLogic GCNConv layer computes the same as PyG GCNConv layer"""

    edge_index = torch.tensor([[0, 1], [1, 0], [1, 2], [2, 1]], dtype=torch.long)
    x = torch.randn(
        (
            3,
            input_size,
        ),
        dtype=torch.float,
    )
    e = torch.ones(
        (4,),
        dtype=torch.float,
    )

    data = Data(x=x, edge_index=edge_index.t().contiguous(), edge_attr=e)
    gcn = GCNConv(input_size, output_size, bias=False)

    template = Template()
    template += neuralogic.nn.module.GCNConv(input_size, output_size, "h", "f", "e")

    model = template.build(
        Settings(chain_pruning=False, iso_value_compression=False, optimizer=Adam(lr=0.001), error_function=MSE())
    )

    parameters = model.parameters()
    torch_parameters = [parameter.tolist() for parameter in gcn.parameters()]

    parameters["weights"][1] = [torch_parameters[0][i] for i in range(0, output_size)]
    model.load_state_dict(parameters)

    example = [
        R.f(0)[x[0]],
        R.f(1)[x[1]],
        R.f(2)[x[2]],
        R.e(0, 1)[e[0]],
        R.e(1, 0)[e[1]],
        R.e(1, 2)[e[2]],
        R.e(2, 1)[e[3]],
    ]

    dataset = Dataset([Sample(R.h(i), example) for i in range(3)])
    bd = model.build_dataset(dataset)

    output = gcn(x=data.x, edge_index=data.edge_index, edge_weight=data.edge_attr)
    result = model.test(bd)

    assert np.allclose([[float(x) for x in xs] for xs in output], [[float(x) for x in xs] for xs in result], atol=10e-5)



from typing import Iterable

from neuralogic.core.constructs.function import FContainer
from neuralogic.core.constructs.metadata import Metadata


class RuleBody:
    __slots__ = "literals", "metadata"

    def __init__(self, lit1, lit2):
        self.literals = [lit1, lit2]
        self.metadata = None

    def __and__(self, other):
        from neuralogic.core.constructs.relation import BaseRelation

        if isinstance(other, (BaseRelation, FContainer)):
            self.literals.append(other)
            return self
        raise NotImplementedError

    def __str__(self) -> str:
        return ", ".join(atom.to_str() for atom in self.literals)

    def __repr__(self) -> str:
        return self.__str__()

    def __or__(self, other) -> "RuleBody":
        if isinstance(other, Iterable):
            other = Metadata.from_iterable(other)
        elif not isinstance(other, Metadata):
            raise NotImplementedError

        self.metadata = other
        return self


class Rule:
    __slots__ = "head", "body", "metadata"

    def __init__(self, head, body):
        self.head = head
        self.metadata: Metadata | None = None

        if head.function is not None:
            raise NotImplementedError(f"Rule head {head} cannot have a function attached")

        if isinstance(body, RuleBody):
            if body.metadata is not None:
                self._set_metadata(body.metadata)
            body = body.literals

        if not isinstance(body, Iterable):
            body = [body]

        self.body = body

        if not isinstance(self.body, FContainer):
            self.body = list(self.body)

    def _contains_function_container(self):
        for lit in self.body:
            if isinstance(lit, FContainer):
                return True
        return False

    def to_str(self, _: bool = False) -> str:
        return str(self)

    def __str__(self) -> str:
        metadata = "" if self.metadata is None is None else f" {self.metadata}"
        if isinstance(self.body, FContainer):
            return f"{self.head.to_str()} :- {self.body.to_str()}.{metadata}"
        return f"{self.head.to_str()} :- {', '.join(atom.to_str() for atom in self.body)}.{metadata}"

    def __repr__(self) -> str:
        return self.to_str()

    def __and__(self, other) -> "Rule":
        if isinstance(other, Iterable):
            self.body.extend(list(other))
        else:
            self.body.append(other)
        return self

    def __or__(self, other) -> "Rule":
        if isinstance(other, Iterable):
            other = Metadata.from_iterable(other)
        elif not isinstance(other, Metadata):
            raise NotImplementedError

        self._set_metadata(other)
        return self

    def _set_metadata(self, metadata: Metadata):
        if metadata.aggregation is not None and metadata.aggregation.rule_head_dependant():
            metadata = metadata.copy()
            metadata.aggregation = metadata.aggregation.process_head(self.head)

        self.metadata = metadata



from neuralogic.core.constructs.function.concat import ConcatAggregation, ConcatCombination
from neuralogic.core.constructs.function.function import (
    TransformationFunction,
    CombinationFunction,
    AggregationFunction,
)
from neuralogic.core.constructs.function.reshape import Reshape
from neuralogic.core.constructs.function.slice import Slice
from neuralogic.core.constructs.function.softmax import SoftmaxAggregation


class Transformation:
    # Element wise
    SIGMOID: TransformationFunction = TransformationFunction("SIGMOID")
    TANH: TransformationFunction = TransformationFunction("TANH")
    SIGNUM: TransformationFunction = TransformationFunction("SIGNUM")
    RELU: TransformationFunction = TransformationFunction("RELU", namespace="transformation.elementwise.ReLu")
    LEAKY_RELU: TransformationFunction = TransformationFunction(
        "LEAKYRELU", namespace="transformation.elementwise.LeakyReLu"
    )
    LUKASIEWICZ: TransformationFunction = TransformationFunction(
        "LUKASIEWICZ", namespace="transformation.elementwise.LukasiewiczSigmoid"
    )
    EXP: TransformationFunction = TransformationFunction("EXP", namespace="transformation.elementwise.Exponentiation")
    SQRT: TransformationFunction = TransformationFunction("SQRT", namespace="transformation.elementwise.SquareRoot")
    INVERSE: TransformationFunction = TransformationFunction("INVERSE")
    REVERSE: TransformationFunction = TransformationFunction("REVERSE")
    LOG: TransformationFunction = TransformationFunction("LOG", namespace="transformation.elementwise.Logarithm")

    # Transformation
    IDENTITY: TransformationFunction = TransformationFunction("IDENTITY", namespace="transformation.join.{name}")
    TRANSP: TransformationFunction = TransformationFunction("TRANSPOSE", namespace="transformation.join.Transposition")
    SOFTMAX: TransformationFunction = TransformationFunction("SOFTMAX", namespace="transformation.join.{name}")
    SPARSEMAX: TransformationFunction = TransformationFunction("SPARSEMAX", namespace="transformation.join.{name}")
    NORM: TransformationFunction = TransformationFunction("NORM", namespace="transformation.join.Normalization")
    SLICE: Slice = Slice("SLICE")
    RESHAPE: Reshape = Reshape("RESHAPE")


class Combination:
    # Aggregation
    AVG: CombinationFunction = CombinationFunction("AVG", namespace="aggregation.Average")
    MAX: CombinationFunction = CombinationFunction("MAX", can_flatten=True, namespace="aggregation.Maximum")
    MIN: CombinationFunction = CombinationFunction("MIN", can_flatten=True, namespace="aggregation.Minimum")
    SUM: CombinationFunction = CombinationFunction(
        "SUM", operator="+", can_flatten=True, namespace="aggregation.{name}"
    )
    COUNT: CombinationFunction = CombinationFunction("COUNT", namespace="aggregation.{name}")

    # Combination
    PRODUCT: CombinationFunction = CombinationFunction("PRODUCT", operator="@")
    ELPRODUCT: CombinationFunction = CombinationFunction(
        "ELPRODUCT", operator="*", can_flatten=True, namespace="combination.ElementProduct"
    )
    SOFTMAX: CombinationFunction = CombinationFunction("SOFTMAX")
    SPARSEMAX: CombinationFunction = CombinationFunction("SPARSEMAX")
    CROSSSUM: CombinationFunction = CombinationFunction("CROSSSUM")
    CONCAT: ConcatCombination = ConcatCombination("CONCAT")
    COSSIM: CombinationFunction = CombinationFunction("COSSIM")


class Aggregation:
    AVG: AggregationFunction = AggregationFunction("AVG")
    MAX: AggregationFunction = AggregationFunction("MAX")
    MIN: AggregationFunction = AggregationFunction("MIN")
    SUM: AggregationFunction = AggregationFunction("SUM")
    COUNT: AggregationFunction = AggregationFunction("COUNT")
    CONCAT: ConcatAggregation = ConcatAggregation("CONCAT")
    SOFTMAX: SoftmaxAggregation = SoftmaxAggregation("SOFTMAX")


class F:
    # Element wise
    sigmoid: TransformationFunction = Transformation.SIGMOID
    tanh: TransformationFunction = Transformation.TANH
    signum: TransformationFunction = Transformation.SIGNUM
    relu: TransformationFunction = Transformation.RELU
    leaky_relu: TransformationFunction = Transformation.LEAKY_RELU
    lukasiewicz: TransformationFunction = Transformation.LUKASIEWICZ
    exp: TransformationFunction = Transformation.EXP
    sqrt: TransformationFunction = Transformation.SQRT
    inverse: TransformationFunction = Transformation.INVERSE
    reverse: TransformationFunction = Transformation.REVERSE
    log: TransformationFunction = Transformation.LOG

    # Transformation
    identity: TransformationFunction = Transformation.IDENTITY
    transp: TransformationFunction = Transformation.TRANSP
    softmax: TransformationFunction = Transformation.SOFTMAX
    sparsemax: TransformationFunction = Transformation.SPARSEMAX
    norm: TransformationFunction = Transformation.NORM
    slice: Slice = Transformation.SLICE
    reshape: Reshape = Transformation.RESHAPE

    # Combination
    avg: CombinationFunction = Combination.AVG
    max: CombinationFunction = Combination.MAX
    min: CombinationFunction = Combination.MIN
    sum: CombinationFunction = Combination.SUM
    count: CombinationFunction = Combination.COUNT

    product: CombinationFunction = Combination.PRODUCT
    elproduct: CombinationFunction = Combination.ELPRODUCT
    softmax_comb: CombinationFunction = Combination.SOFTMAX
    sparsemax_comb: CombinationFunction = Combination.SPARSEMAX
    crossum: CombinationFunction = Combination.CROSSSUM
    concat: ConcatCombination = Combination.CONCAT
    cossim: CombinationFunction = Combination.COSSIM

    # Aggregation
    avg_agg: AggregationFunction = Aggregation.AVG
    max_agg: AggregationFunction = Aggregation.MAX
    min_agg: AggregationFunction = Aggregation.MIN
    sum_agg: AggregationFunction = Aggregation.SUM
    count_agg: AggregationFunction = Aggregation.COUNT
    concat_agg: ConcatAggregation = Aggregation.CONCAT
    softmax_agg: SoftmaxAggregation = Aggregation.SOFTMAX



from typing import List

from neuralogic.core.constructs.function.function import TransformationFunction, AggregationFunction
from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Transformation, Aggregation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class MetaConv(Module):
    r"""
    Metagraph Convolutional Unit layer from
    `Meta-GNN: metagraph neural network for semi-supervised learning in attributed heterogeneous information networks <https://dl.acm.org/doi/10.1145/3341161.3342859>`_.
    Which can be expressed as:

    .. math::
        \mathbf{x}^{\prime}_i = act(\mathbf{W_0} \cdot \mathbf{x}_i + {agg}_{j \in \mathcal{N}_r(i)}
        \sum_{k \in \mathcal{K}}
        (\mathbf{W_k} \cdot \mathbf{x}_j))

    Where *act* is an activation function, *agg* aggregation function (by default average), :math:`W_0` is a learnable
    root parameter and :math:`W_k` is a learnable parameter for each role.

    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    role_name : str | None
        Role predicate name to use for role relations. When :code:`None`, elements from :code:`roles` are used instead.
    roles : List[str]
        List of relations' names
    activation : TransformationFunction
        Activation function of the output.
        Default: ``Transformation.SIGMOID``
    aggregation : AggregationFunction
        Aggregation function of nodes' neighbors.
        Default: ``Aggregation.AVG``

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        role_name: str | None,
        roles: List[str],
        activation: TransformationFunction = Transformation.SIGMOID,
        aggregation: AggregationFunction = Aggregation.AVG,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.role_name = role_name

        self.roles = roles

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.aggregation = aggregation
        self.activation = activation

    def __call__(self):
        head = R.get(self.output_name)(V.I)
        role_head = R.get(f"{self.output_name}__roles")

        metadata = Metadata(aggregation=self.aggregation)
        feature = R.get(self.feature_name)(V.J)[self.out_channels, self.in_channels]

        if self.role_name is not None:
            role_rules = [
                (role_head(V.I, role) <= (feature, R.get(self.role_name)(V.J, role, V.I))) for role in self.roles
            ]
        else:
            role_rules = [(role_head(V.I, role) <= (feature, R.get(role)(V.J, V.I))) for role in self.roles]

        return [
            (head <= role_head(V.I, V.R)) | metadata,
            (head <= R.get(self.feature_name)(V.I)[self.out_channels, self.in_channels]) | metadata,
            *role_rules,
            R.get(self.output_name) / 1 | Metadata(transformation=self.activation),
        ]



import enum
from typing import Any, Dict, Optional

from neuralogic.core.constructs.java_objects import ValueFactory
from neuralogic.utils.visualize import draw_sample, draw_grounding


class NeuronType(enum.StrEnum):
    Aggregation = "AggregationNeuron"
    Atom = "AtomNeuron"
    Negation = "NegationNeuron"
    Rule = "RuleNeuron"
    SplittableAggregation = "SplittableAggregationNeuron"
    WeightedAtom = "WeightedAtomNeuron"
    WeightedRule = "WeightedRuleNeuron"
    Fact = "FactNeuron"


class Atom:
    __slots__ = "substitutions", "_atom", "_predicate", "_arity"

    def __init__(self, atom, substitutions: Dict):
        self.substitutions = substitutions
        self._atom = atom

        self._predicate = atom.predicateName()
        self._arity = atom.arity()

    @property
    def predicate(self):
        return self._predicate

    @property
    def arity(self):
        return self._arity

    def node_type(self) -> NeuronType:
        return NeuronType(self._atom.getClass().getSimpleName())

    def __str__(self):
        return str(self._atom)


class Neuron(Atom):
    def __init__(self, neuron, substitutions: Dict):
        self.substitutions = substitutions
        self._atom = neuron

        self._predicate = neuron.getName()
        self._arity = len(substitutions)

    @property
    def value(self):
        return ValueFactory.from_java(self._atom.getRawState().getValue())

    @property
    def gradient(self):
        return ValueFactory.from_java(self._atom.getRawState().getGradient())


class NeuralSample:
    __slots__ = "_java_sample", "_neurons"

    def __init__(self, sample):
        self._java_sample = sample
        self._neurons = None

    @property
    def neurons(self):
        if self._neurons is None:
            self._neurons = self._get_neurons()
        return self._neurons

    @property
    def target(self):
        return ValueFactory.from_java(self._java_sample.target)

    def get_neurons(self, literal, neuron_type: NeuronType | None = NeuronType.Atom):
        literal_name = literal.predicate.name
        literal_arity = literal.predicate.arity

        neurons = []
        neurons_by_name_vals = []

        if neuron_type is None:
            neurons_by_name_vals = self.neurons.values()
        elif neuron_type in self.neurons:
            neurons_by_name_vals = [self.neurons[neuron_type]]

        for nodes_by_name in neurons_by_name_vals:
            if literal_name not in nodes_by_name:
                continue

            for subs, value in nodes_by_name[literal_name].items():
                if len(subs) != literal_arity:
                    continue

                literal_subs = {}
                for term, sub in zip(literal.terms, subs):
                    term_str = str(term)

                    if term_str[0] == term_str[0].upper() and term_str[0] != term_str[0].lower():
                        if term_str in literal_subs and sub != literal_subs[term_str]:
                            break
                        literal_subs[str(term)] = sub
                        continue

                    if str(term) != sub:
                        break
                else:
                    neurons.append(Neuron(value, literal_subs))
        return neurons

    def _get_neurons(self):
        nodes = {}

        for node in self._java_sample.query.evidence.allNeuronsTopologic:
            node_type = str(node.getClass().getSimpleName())
            name = str(node.name).strip()

            bracket = name.rfind("(")
            space = name.rfind(" ", 0, bracket if bracket != -1 else None)

            substitutions = tuple()
            if bracket != -1:
                subs = name[bracket + 1 :]
                name = name[space + 1 : bracket]

                r_bracket = subs.find(")")
                substitutions = tuple(subs[:r_bracket].split(", "))
            elif space != -1:
                name = name[space + 1 :]

            if node_type not in nodes:
                nodes[node_type] = {}

            if name not in nodes[node_type]:
                nodes[node_type][name] = {}

            nodes[node_type][name][substitutions] = node
        return nodes

    def get_fact(self, fact):
        for term in fact.terms:
            term_str = str(term)

            if term_str[0] == term_str[0].upper() and term_str[0] != term_str[0].lower():
                raise ValueError(f"{fact} is not a fact")

        return self.get_neurons(fact, NeuronType.Fact)

    def set_fact_value(self, fact, value) -> int:
        for term in fact.terms:
            term_str = str(term)

            if term_str[0] == term_str[0].upper() and term_str[0] != term_str[0].lower():
                raise ValueError(f"{fact} is not a fact")

        node = self.get_neurons(fact, NeuronType.Fact)

        if len(node) == 0:
            return -1

        sample_fact = node[0]._atom
        sample_fact.getRawState().setValue(value)
        sample_fact.offset.value = value

        return sample_fact.index

    def draw(
        self,
        filename: str | None = None,
        show=True,
        img_type="png",
        value_detail: int = 0,
        graphviz_path: str | None = None,
        *args,
        **kwargs,
    ):
        return draw_sample(self, filename, show, img_type, value_detail, graphviz_path, *args, **kwargs)

    def __str__(self):
        return str(self._java_sample)


class Grounding:
    __slots__ = ("_grounding", "_atoms")

    def __init__(self, grounding):
        self._grounding = grounding
        self._atoms = None

    def draw(
        self,
        filename: Optional[str] = None,
        show=True,
        img_type="png",
        value_detail: int = 0,
        graphviz_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        return draw_grounding(self._grounding, filename, show, img_type, value_detail, graphviz_path, *args, **kwargs)

    def __hash__(self):
        return hash(self._grounding)

    def __eq__(self, other):
        if not isinstance(other, Grounding):
            return False
        return other._grounding == self._grounding

    def __str__(self):
        return str(self._grounding)

    @property
    def atoms(self) -> dict[str, dict[tuple[str, ...], Any]]:
        if self._atoms is None:
            self._atoms = self._get_atoms()
        return self._atoms

    def get_atoms(self, literal) -> list[Atom]:
        literal_name = literal.predicate.name
        literal_arity = literal.predicate.arity

        if literal_name not in self.atoms:
            return []

        nodes = []
        for subs, value in self.atoms[literal_name].items():
            if len(subs) != literal_arity:
                continue

            literal_subs = {}
            for term, sub in zip(literal.terms, subs):
                term_str = str(term)

                if term_str[0] == term_str[0].upper() and term_str[0] != term_str[0].lower():
                    if term_str in literal_subs and sub != literal_subs[term_str]:
                        break
                    literal_subs[str(term)] = sub
                    continue

                if str(term) != sub:
                    break
            else:
                nodes.append(Atom(value, literal_subs))
        return nodes

    def _get_atoms(self):
        atoms = {}

        for literal in self._grounding.groundingWrap.getGroundTemplate().derivedGroundFacts:
            self._process_literal(literal, atoms)

        for literal in self._grounding.groundingWrap.getGroundTemplate().groundFacts:
            self._process_literal(literal, atoms)

        return atoms

    def _process_literal(self, literal, atoms):
        name = str(literal).strip()

        bracket = name.rfind("(")
        space = name.rfind(" ", 0, bracket if bracket != -1 else None)

        substitutions = tuple()
        if bracket != -1:
            subs = name[bracket + 1 :]
            name = name[space + 1 : bracket]

            r_bracket = subs.find(")")
            substitutions = tuple(subs[:r_bracket].split(", "))
        elif space != -1:
            name = name[space + 1 :]

        if name not in atoms:
            atoms[name] = {}

        atoms[name][substitutions] = literal



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Transformation, Aggregation, Combination
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class GINEConv(Module):
    r"""
    GINEConv layer from `"Strategies for Pre-training Graph Neural Networks" <https://arxiv.org/abs/1905.12265>`_.

    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    feature_name : str
        Feature predicate name to get features from.
    edge_name : str
        Edge predicate name to use for neighborhood relations.
    nn_name : str
        Neural network predicate name.
    eps : float
        :math:`\epsilon`-value.
        Default: ``0.0``
    train_eps : bool
        Is ``eps`` trainable parameter.
        Default: ``false``
    edge_dim : int | None
        Dimension of edge features (``None`` is projection to ``in_channels`` is not needed).
        Default: ``None``
    """

    def __init__(
        self,
        in_channels: int,
        feature_name: str,
        edge_name: str,
        nn_name: str,
        eps: float = 0.0,
        train_eps: bool = False,
        edge_dim: int | None = None,
    ):
        self.feature_name = feature_name
        self.edge_name = edge_name
        self.nn_name = nn_name

        self.eps = eps
        self.train_eps = train_eps

        self.in_channels = in_channels
        self.edge_dim: int | None = edge_dim

    def __call__(self):
        feat_sum = R.get(f"{self.nn_name}__gine_feat_sum")
        feat_agg = R.get(f"{self.nn_name}__gine_feat_agg")
        out = R.get(f"{self.nn_name}__gine_out")
        x = R.get(self.feature_name)
        e = R.get(self.edge_name)

        x_eps = x(V.I)[self.eps + 1]
        if not self.train_eps:
            x_eps = x_eps.fixed()

        e_proj = []
        if self.edge_dim is not None:
            e = R.get(f"{self.nn_name}__gine_edge_proj")
            e_proj = [e(V.I, V.J)[self.in_channels, self.edge_dim] <= R.get(self.edge_name)(V.I, V.J)]

        return [
            *e_proj,
            (feat_sum(V.I, V.J) <= (x(V.J), e(V.J, V.I)))
            | Metadata(transformation=Transformation.RELU, combination=Combination.SUM),
            (feat_agg(V.I) <= feat_sum(V.I, V.J)) | Metadata(aggregation=Aggregation.SUM),
            (out(V.I) <= (x_eps, feat_agg(V.I))) | Metadata(combination=Combination.SUM),
            R.get(self.nn_name)(V.I) <= out(V.I),
        ]



import jpype

from neuralogic.core.constructs.function.function import Function


class FunctionGraph(Function):
    __slots__ = ("function_graph",)

    def __init__(
        self,
        name: str,
        *,
        function_graph,
    ):
        super().__init__(name)
        self.function_graph = function_graph

    def is_parametrized(self) -> bool:
        return True

    def get(self):
        return jpype.JClass("cz.cvut.fel.ida.algebra.functions.combination.FunctionGraph")(
            self.name, self.function_graph
        )

    def __str__(self):
        return self.name



from typing import Collection

import jpype

from neuralogic.setup import is_initialized, initialize
from neuralogic.core.constructs.java_objects import ValueFactory
from neuralogic.core.builder import DatasetBuilder
from neuralogic.core.builder.dataset import BuiltDataset, GroundedDataset
from neuralogic.core.settings.settings_proxy import SettingsProxy
from neuralogic.dataset import Dataset
from neuralogic.dataset.base import BaseDataset

from neuralogic.utils.visualize import draw_model


Value = list | float


class NeuralModule:
    def __init__(self):
        if not is_initialized():
            initialize()

        self._need_sync = False
        self._value_factory = ValueFactory()

        self._parsed_template = None
        self._dataset_builder: DatasetBuilder | None = None
        self._settings: SettingsProxy | None = None

        self._neural_model = None
        self._strategy = None
        self._trainer = None

        self._invalidation = None
        self._evaluation = None
        self._backpropagation = None

        self._weight_updater = None
        self._tensor_parameters = None
        self._torch_module = None

    def ground(
        self,
        dataset: BaseDataset,
        *,
        batch_size: int = 1,
        learnable_facts: bool = False,
        progress: bool = False,
    ) -> GroundedDataset:
        if self._dataset_builder is None or self._settings is None:
            raise ValueError("template is not built")

        return self._dataset_builder.ground_dataset(
            dataset,
            self._settings,
            batch_size=batch_size,
            learnable_facts=learnable_facts,
            progress=progress,
        )

    def build_dataset(
        self,
        dataset: BaseDataset | GroundedDataset,
        *,
        batch_size: int = 1,
        learnable_facts: bool = False,
        progress: bool = False,
    ) -> BuiltDataset:
        if self._dataset_builder is None or self._settings is None:
            raise ValueError("template is not built")

        return self._dataset_builder.build_dataset(
            dataset,
            self._settings,
            batch_size=batch_size,
            learnable_facts=learnable_facts,
            progress=progress,
        )

    def __call__(self, dataset=None):
        samples, _ = self._dataset_to_samples(dataset)
        sample_collection = samples if isinstance(samples, Collection) else [samples]

        for sample in sample_collection:
            self._trainer.invalidateSample(self._invalidation, sample._java_sample)

        results = [
            self._value_factory.from_java(
                self._trainer.evaluateSample(self._evaluation, sample._java_sample).getOutput(),
            )
            for sample in sample_collection
        ]

        if self._torch_module is None:
            return results

        return self._torch_module.forward(self, samples, results)

    def forward(self, dataset):
        return self(dataset)

    def train(self, dataset, epochs: int = 1) -> Value:
        samples, batch_size = self._dataset_to_samples(dataset)

        if not isinstance(samples, Collection):
            result = self._strategy.learnSample(samples._java_sample)
            res = (
                ValueFactory.from_java(result.getTarget()),
                ValueFactory.from_java(result.getOutput()),
                ValueFactory.from_java(result.errorValue()),
            )
        else:
            sample_array = jpype.java.util.ArrayList([sample._java_sample for sample in samples])
            results = self._strategy.learnSamples(sample_array, epochs, batch_size)
            res = [
                (
                    ValueFactory.from_java(result.getTarget()),
                    ValueFactory.from_java(result.getOutput()),
                    ValueFactory.from_java(result.errorValue()),
                ) for result in results
            ]

        self._update_tensor_parameters()
        return res

    def test(self, dataset) -> Value:
        samples, batch_size = self._dataset_to_samples(dataset)

        if not isinstance(samples, Collection):
            return ValueFactory.from_java(self._strategy.evaluateSample(samples._java_sample))

        sample_array = jpype.java.util.ArrayList([sample._java_sample for sample in samples])
        results = self._strategy.evaluateSamples(sample_array, batch_size)

        return [ValueFactory.from_java(result) for result in results]

    def reset_parameters(self):
        self._strategy.resetParameters()

    def parameters(self) -> dict:
        return self.state_dict()

    def state_dict(self) -> dict:
        weights = self._neural_model.getAllWeights()
        weights_dict = {}
        weight_names = {}

        for weight in weights:
            if weight.isLearnable:
                weights_dict[weight.index] = ValueFactory.from_java(weight.value)
                weight_names[weight.index] = str(weight.name)
        return {
            "weights": weights_dict,
            "weight_names": weight_names,
        }

    def tensor_parameters(self):
        if self._torch_module is None:
            raise NotImplementedError

        self._tensor_parameters = self._torch_module.tensor_parameters(
            self._tensor_parameters,
            self._weight_updater,
            self._value_factory,
            self._neural_model,
        )

        return list(self._tensor_parameters)

    def _update_tensor_parameters(self):
        if self._torch_module is not None:
            self._torch_module.update_tensor_parameters(self._tensor_parameters)

    def load_state_dict(self, state_dict: dict):
        self._sync_template(state_dict, self._neural_model.getAllWeights())

        if self._torch_module is not None:
            self._torch_module.update_tensor_parameters(self._tensor_parameters)

    def draw(
        self,
        filename: str | None = None,
        show=True,
        img_type="png",
        value_detail: int = 0,
        graphviz_path: str | None = None,
        *args,
        **kwargs,
    ):
        if self._dataset_builder is None or self._settings is None:
            raise ValueError("template is not built")
        return draw_model(self, filename, show, img_type, value_detail, graphviz_path, *args, **kwargs)

    def _initialize_neural_module(self, dataset_builder: DatasetBuilder, settings: SettingsProxy, model, torch: bool):
        self._dataset_builder = dataset_builder
        self._settings = settings
        self._neural_model = model

        if torch:
            try:
                import torch
            except:
                raise Exception("torch is not installed in the environment")

            from neuralogic.core.torch.neural_module import TorchNeuralModule

            self._torch_module = TorchNeuralModule()

        optimizer = self._settings.optimizer.initialize()
        lr_decay = self._settings.optimizer.get_lr_decay()

        python_strategy = jpype.JClass(
            "cz.cvut.fel.ida.neural.networks.computation.training.strategies.PythonTrainingStrategy"
        )

        self._strategy = python_strategy(settings.settings, model, optimizer, lr_decay)
        self._trainer = self._strategy.getTrainer()

        self._invalidation = self._trainer.getInvalidation()
        self._evaluation = self._trainer.getEvaluation()
        self._backpropagation = self._trainer.getBackpropagation()
        self._weight_updater = self._backpropagation.weightUpdater

        self.reset_parameters()

    def _dataset_to_samples(self, dataset):
        if isinstance(dataset, Dataset):
            dataset = self.build_dataset(dataset)
            return dataset._samples, dataset._batch_size

        if isinstance(dataset, GroundedDataset):
            dataset = dataset.neuralize()
            return dataset._samples, dataset._batch_size

        if isinstance(dataset, BuiltDataset):
            return dataset._samples, dataset._batch_size
        return dataset, 1

    def _sync_template(self, state_dict: dict | None = None, weights=None):
        state_dict = self.state_dict() if state_dict is None else state_dict
        weights = self._parsed_template.getAllWeights() if weights is None else weights
        weight_dict = state_dict["weights"]

        for weight in weights:
            if not weight.isLearnable:
                continue
            weight_value = weight.value

            index = weight.index
            value = weight_dict[index]

            if isinstance(value, (float, int)):
                weight_value.set(0, float(value))
                continue

            if isinstance(value[0], (float, int)):
                for i, val in enumerate(value):
                    weight_value.set(i, float(val))
                continue

            cols = len(value[0])

            for i, values in enumerate(value):
                for j, val in enumerate(values):
                    weight_value.set(i * cols + j, float(val))

    def _backprop(self, sample, gradient):
        _, gradient_value = self._value_factory.get_value(gradient)

        weight_updater = self._backpropagation.backpropagate(sample._java_sample, gradient_value)
        state_index = self._backpropagation.backproper

        return state_index, weight_updater



import numpy as np
import pytest
import torch

from neuralogic.core import Template, Settings, R, V, Transformation
from neuralogic.dataset import Dataset, Sample
from neuralogic.nn.loss import MSE

from neuralogic.nn.module import GRU, RNN, LSTM
from neuralogic.optim import Adam, SGD


@pytest.mark.parametrize(
    "input_size, hidden_size, sequence_len, epochs",
    [
        (10, 5, 10, 500),
    ],
)
def test_gru_module(input_size, hidden_size, sequence_len, epochs):
    """Test that PyNeuraLogic GRU layer computes the same as PyTorch GRU layer (with backprop)"""
    torch_input = torch.randn((sequence_len, input_size))
    h0 = torch.randn((1, hidden_size))
    target = torch.randn((hidden_size,))

    rnn = torch.nn.GRU(input_size, hidden_size, 1, bias=False)

    template = Template()
    template += GRU(input_size, hidden_size, "h", "f", "h0", arity=0)

    model = template.build(
        Settings(chain_pruning=False, iso_value_compression=False, optimizer=Adam(lr=0.001), error_function=MSE())
    )

    parameters = model.parameters()
    torch_parameters = [parameter.tolist() for parameter in rnn.parameters()]

    parameters["weights"][0] = [torch_parameters[0][i] for i in range(0, hidden_size)]
    parameters["weights"][2] = [torch_parameters[0][i] for i in range(1 * hidden_size, 1 * hidden_size + hidden_size)]
    parameters["weights"][5] = [torch_parameters[0][i] for i in range(2 * hidden_size, 2 * hidden_size + hidden_size)]

    parameters["weights"][1] = [torch_parameters[1][i] for i in range(0, hidden_size)]
    parameters["weights"][3] = [torch_parameters[1][i] for i in range(1 * hidden_size, 1 * hidden_size + hidden_size)]
    parameters["weights"][4] = [torch_parameters[1][i] for i in range(2 * hidden_size, 2 * hidden_size + hidden_size)]

    model.load_state_dict(parameters)

    dataset = Dataset(
        [
            Sample(
                R.h(sequence_len)[target.detach().numpy().tolist()],
                [
                    R.h0[[float(h) for h in h0[0]]],
                    *[R.f(i + 1)[[float(h) for h in torch_input[i]]] for i in range(sequence_len)],
                ],
            )
        ]
    )

    bd = model.build_dataset(dataset)

    optimizer = torch.optim.Adam(rnn.parameters(), lr=0.001)
    loss_fun = torch.nn.MSELoss()

    for _ in range(epochs):
        output, _ = rnn(torch_input, h0)
        loss = loss_fun(output[-1], target)

        optimizer.zero_grad(set_to_none=True)
        loss.backward()
        optimizer.step()

        result = model.train(bd)
        assert np.allclose([float(x) for x in output[-1]], [float(x) for x in result[0][1]], atol=10e-5)


@pytest.mark.parametrize(
    "input_size, hidden_size, sequence_len, epochs",
    [
        (10, 5, 10, 500),
    ],
)
def test_rnn_module(input_size, hidden_size, sequence_len, epochs):
    """Test that PyNeuraLogic RNN layer computes the same as PyTorch RNN layer (with backprop)"""
    torch_input = torch.randn((sequence_len, input_size))
    h0 = torch.randn((1, hidden_size))
    target = torch.randn((hidden_size,))

    rnn = torch.nn.RNN(input_size, hidden_size, 1, bias=False)

    template = Template()
    template += RNN(input_size, hidden_size, "h", "f", "h0", arity=0)

    model = template.build(
        Settings(chain_pruning=False, iso_value_compression=False, optimizer=Adam(lr=0.001), error_function=MSE())
    )

    parameters = model.parameters()
    torch_parameters = [parameter.tolist() for parameter in rnn.parameters()]

    parameters["weights"][0] = torch_parameters[0]
    parameters["weights"][1] = torch_parameters[1]

    model.load_state_dict(parameters)

    dataset = Dataset(
        [
            Sample(
                R.h(sequence_len)[target.detach().numpy().tolist()],
                [
                    R.h0[[float(h) for h in h0[0]]],
                    *[R.f(i + 1)[[float(h) for h in torch_input[i]]] for i in range(sequence_len)],
                ],
            ),
        ]
    )

    bd = model.build_dataset(dataset)

    optimizer = torch.optim.Adam(rnn.parameters(), lr=0.001)
    loss_fun = torch.nn.MSELoss()

    for _ in range(epochs):
        output, _ = rnn(torch_input, h0)
        loss = loss_fun(output[-1], target)

        optimizer.zero_grad(set_to_none=True)
        loss.backward()
        optimizer.step()

        result = model.train(bd)
        assert np.allclose([float(x) for x in output[-1]], [float(x) for x in result[0][1]], atol=10e-5)


@pytest.mark.parametrize(
    "input_size, hidden_size, sequence_len, epochs",
    [
        (10, 5, 10, 500),
    ],
)
def test_lstm_module(input_size, hidden_size, sequence_len, epochs):
    """Test that PyNeuraLogic LSTM layer computes the same as PyTorch LSTM layer (with backprop)"""
    torch_input = torch.randn((sequence_len, input_size))
    h0 = torch.randn((1, hidden_size))
    c0 = torch.randn((1, hidden_size))
    target = torch.randn((hidden_size,))

    rnn = torch.nn.LSTM(input_size, hidden_size, 1, bias=False)

    template = Template()
    template += LSTM(input_size, hidden_size, "h", "f", "h0", "c0", arity=0)

    model = template.build(
        Settings(chain_pruning=False, iso_value_compression=False, optimizer=Adam(lr=0.001), error_function=MSE())
    )

    parameters = model.parameters()
    torch_parameters = [parameter.tolist() for parameter in rnn.parameters()]

    parameters["weights"][0] = [torch_parameters[0][i] for i in range(0, hidden_size)]
    parameters["weights"][2] = [torch_parameters[0][i] for i in range(1 * hidden_size, 1 * hidden_size + hidden_size)]
    parameters["weights"][4] = [torch_parameters[0][i] for i in range(3 * hidden_size, 3 * hidden_size + hidden_size)]
    parameters["weights"][6] = [torch_parameters[0][i] for i in range(2 * hidden_size, 2 * hidden_size + hidden_size)]

    parameters["weights"][1] = [torch_parameters[1][i] for i in range(0, hidden_size)]
    parameters["weights"][3] = [torch_parameters[1][i] for i in range(1 * hidden_size, 1 * hidden_size + hidden_size)]
    parameters["weights"][5] = [torch_parameters[1][i] for i in range(3 * hidden_size, 3 * hidden_size + hidden_size)]
    parameters["weights"][7] = [torch_parameters[1][i] for i in range(2 * hidden_size, 2 * hidden_size + hidden_size)]

    model.load_state_dict(parameters)

    dataset = Dataset(
        [
            Sample(
                R.h(sequence_len)[target.detach().numpy().tolist()],
                [
                    R.c0[[float(c) for c in c0[0]]],
                    R.h0[[float(h) for h in h0[0]]],
                    *[R.f(i + 1)[[float(h) for h in torch_input[i]]] for i in range(sequence_len)],
                ],
            ),
        ]
    )

    bd = model.build_dataset(dataset)

    optimizer = torch.optim.Adam(rnn.parameters(), lr=0.001)
    loss_fun = torch.nn.MSELoss()

    for _ in range(epochs):
        output, _ = rnn(torch_input, (h0, c0))
        loss = loss_fun(output[-1], target)

        optimizer.zero_grad(set_to_none=True)
        loss.backward()
        optimizer.step()

        result = model.train(bd)
        assert np.allclose([float(x) for x in output[-1]], [float(x) for x in result[0][1]], atol=10e-5)


@pytest.mark.parametrize(
    "input_size, hidden_size, sequence_len, epochs",
    [
        (5, 5, 10, 500),
    ],
)
def test_rnn_module_with_pytorch(input_size, hidden_size, sequence_len, epochs):
    """Test that PyNeuraLogic RNN layer computes the same as PyTorch RNN layer (with backprop)"""
    torch_input = torch.randn((sequence_len, input_size))
    h0 = torch.randn((1, hidden_size))
    target = torch.randn((hidden_size,))

    rnn = torch.nn.RNN(input_size, hidden_size, 1, bias=False)

    template = Template()
    template += RNN(input_size, hidden_size, "h", "f", "h0", arity=0)

    model = template.build(Settings(chain_pruning=False, iso_value_compression=False, error_function=MSE()), torch=True)

    parameters = model.parameters()
    pyneuralogic_tensor_parameters = model.tensor_parameters()

    torch_parameters = [parameter.tolist() for parameter in rnn.parameters()]

    parameters["weights"][0] = torch_parameters[0]
    parameters["weights"][1] = torch_parameters[1]

    model.load_state_dict(parameters)

    dataset = Dataset(
        [
            Sample(
                R.h(sequence_len)[target.detach().numpy().tolist()],
                [
                    R.h0[[float(h) for h in h0[0]]],
                    *[R.f(i + 1)[[float(h) for h in torch_input[i]]] for i in range(sequence_len)],
                ],
            ),
        ]
    )

    bd = model.build_dataset(dataset)

    optimizer = torch.optim.Adam(rnn.parameters(), lr=0.001)
    pynelo_torch_optim = torch.optim.Adam(params=pyneuralogic_tensor_parameters, lr=0.001)
    loss_fun = torch.nn.MSELoss()
    pynelo_loss_fun = torch.nn.MSELoss()

    for _ in range(epochs):
        output, _ = rnn(torch_input, h0)
        loss = loss_fun(output[-1], target)

        optimizer.zero_grad(set_to_none=True)
        loss.backward()
        optimizer.step()

        result = model(bd)
        pynelo_loss = pynelo_loss_fun(result[-1], target)

        pynelo_loss.backward()
        pynelo_torch_optim.step()

        assert np.allclose([float(x) for x in output[-1]], [float(x) for x in result[-1]], atol=10e-5)


@pytest.mark.parametrize(
    "input_size, hidden_size, sequence_len, epochs",
    [
        (1, 1, 2, 5),
    ],
)
def test_rnn_custom(input_size, hidden_size, sequence_len, epochs):
    """
    Test alignment on a manually crafted example that's easier to precisely debug on a piece of paper
        THIS IS IN THE OLD SYNTAX, NEEDS REWRITING TO THE NEW ONE:)
    """

    torch_input = torch.tensor([[1.0], [2.0]])  # manual input setup
    h0 = torch.tensor([[0.0]])
    target = torch.tensor([[1.0]])

    rnn = torch.nn.RNN(input_size, hidden_size, 1, bias=False, nonlinearity="relu")

    template = Template()
    template += RNN(input_size, hidden_size, "h", "f", "h0", arity=0, activation=Transformation.RELU)

    model = template.build(
        Settings(chain_pruning=False, iso_value_compression=False, optimizer=SGD(lr=0.001), error_function=MSE())
    )

    parameters = model.parameters()
    for parameter in rnn.parameters():
        parameter.data = torch.tensor([[0.5]])  # manual weight setup
    torch_parameters = [parameter.tolist() for parameter in rnn.parameters()]

    parameters["weights"][0] = torch_parameters[0]
    parameters["weights"][1] = torch_parameters[1]
    model.load_state_dict(parameters)

    dataset = Dataset(
        [
            Sample(
                R.h(sequence_len)[target.detach().numpy().tolist()],
                [
                    R.h0[[float(h) for h in h0[0]]],
                    *[R.f(i + 1)[[float(h) for h in torch_input[i]]] for i in range(sequence_len)],
                ],
            ),
        ]
    )

    bd = model.build_dataset(dataset)

    optimizer = torch.optim.SGD(rnn.parameters(), lr=0.001)
    loss_fun = torch.nn.MSELoss()

    for _ in range(epochs):
        output, _ = rnn(torch_input, h0)
        loss = loss_fun(output[-1], target)

        optimizer.zero_grad(set_to_none=True)
        loss.backward()
        optimizer.step()

        result  = model.train(bd, epochs=1)
        assert np.allclose([float(x) for x in output[-1]], [float(x) for x in result[0][1]], atol=10e-5)


@pytest.mark.parametrize(
    "input_size, hidden_size, sequence_len, epochs",
    [
        (10, 5, 10, 500),
    ],
)
def test_lstm_module_simple(input_size, hidden_size, sequence_len, epochs):
    """
    Test that PyNeuraLogic LSTM layer computes the same as PyTorch LSTM layer (with backprop).
    """
    torch_input = torch.randn((sequence_len, input_size))
    h0 = torch.randn((1, hidden_size))
    c0 = torch.randn((1, hidden_size))
    target = torch.randn((hidden_size,))

    rnn = torch.nn.LSTM(input_size, hidden_size, 1, bias=False)

    template = Template()

    template += [
        R.h(0) <= R.h0,
        R.h__c(0) <= R.c0,
        (
            R.h__i(V.T)
            <= R.f(V.T)[hidden_size, input_size] + R.h(V.Z)[hidden_size, hidden_size] + R.special.next(V.Z, V.T)
        )
        | [Transformation.SIGMOID],
        (
            R.h__f(V.T)
            <= R.f(V.T)[hidden_size, input_size] + R.h(V.Z)[hidden_size, hidden_size] + R.special.next(V.Z, V.T)
        )
        | [Transformation.SIGMOID],
        (
            R.h__o(V.T)
            <= R.f(V.T)[hidden_size, input_size] + R.h(V.Z)[hidden_size, hidden_size] + R.special.next(V.Z, V.T)
        )
        | [Transformation.SIGMOID],
        (
            R.h__n(V.T)
            <= R.f(V.T)[hidden_size, input_size] + R.h(V.Z)[hidden_size, hidden_size] + R.special.next(V.Z, V.T)
        )
        | [Transformation.TANH],
        R.h__c(V.T) <= R.h__f(V.T) * R.h__c(V.Z) + R.h__i(V.T) * R.h__n(V.T) + R.special.next(V.Z, V.T),
        R.h(V.T) <= R.h__o(V.T) * Transformation.TANH(R.h__c(V.T)),
    ]

    model = template.build(
        Settings(chain_pruning=False, iso_value_compression=False, optimizer=Adam(lr=0.001), error_function=MSE())
    )

    parameters = model.parameters()
    torch_parameters = [parameter.tolist() for parameter in rnn.parameters()]

    parameters["weights"][0] = [torch_parameters[0][i] for i in range(0, hidden_size)]
    parameters["weights"][2] = [torch_parameters[0][i] for i in range(1 * hidden_size, 1 * hidden_size + hidden_size)]
    parameters["weights"][4] = [torch_parameters[0][i] for i in range(3 * hidden_size, 3 * hidden_size + hidden_size)]
    parameters["weights"][6] = [torch_parameters[0][i] for i in range(2 * hidden_size, 2 * hidden_size + hidden_size)]

    parameters["weights"][1] = [torch_parameters[1][i] for i in range(0, hidden_size)]
    parameters["weights"][3] = [torch_parameters[1][i] for i in range(1 * hidden_size, 1 * hidden_size + hidden_size)]
    parameters["weights"][5] = [torch_parameters[1][i] for i in range(3 * hidden_size, 3 * hidden_size + hidden_size)]
    parameters["weights"][7] = [torch_parameters[1][i] for i in range(2 * hidden_size, 2 * hidden_size + hidden_size)]

    model.load_state_dict(parameters)

    dataset = Dataset(
        [
            Sample(
                R.h(sequence_len)[target.detach().numpy().tolist()],
                [
                    R.c0[[float(c) for c in c0[0]]],
                    R.h0[[float(h) for h in h0[0]]],
                    *[R.f(i + 1)[[float(h) for h in torch_input[i]]] for i in range(sequence_len)],
                ],
            ),
        ]
    )

    bd = model.build_dataset(dataset)

    optimizer = torch.optim.Adam(rnn.parameters(), lr=0.001)
    loss_fun = torch.nn.MSELoss()

    for _ in range(epochs):
        output, _ = rnn(torch_input, (h0, c0))
        loss = loss_fun(output[-1], target)

        optimizer.zero_grad(set_to_none=True)
        loss.backward()
        optimizer.step()

        result = model.train(bd, 1)
        assert np.allclose([float(x) for x in output[-1]], [float(x) for x in result[0][1]], atol=10e-5)


@pytest.mark.parametrize(
    "input_size, hidden_size, sequence_len, epochs",
    [
        (10, 5, 10, 500),
    ],
)
def test_gru_module_simple(input_size, hidden_size, sequence_len, epochs):
    """Test that PyNeuraLogic GRU layer computes the same as PyTorch GRU layer (with backprop)"""
    torch_input = torch.randn((sequence_len, input_size))
    h0 = torch.randn((1, hidden_size))
    target = torch.randn((hidden_size,))

    rnn = torch.nn.GRU(input_size, hidden_size, 1, bias=False)

    template = Template()

    template += R.h(0) <= R.h0
    template += (
        R.h__r(V.T) <= R.f(V.T)[hidden_size, input_size] + R.h(V.Z)[hidden_size, hidden_size] + R.special.next(V.Z, V.T)
    ) | [Transformation.SIGMOID]
    template += (
        R.h__z(V.T) <= R.f(V.T)[hidden_size, input_size] + R.h(V.Z)[hidden_size, hidden_size] + R.special.next(V.Z, V.T)
    ) | [Transformation.SIGMOID]
    template += (
        R.h__n(V.T)
        <= (R.h__r(V.T) * R.h(V.Z)[hidden_size, hidden_size])
        + R.f(V.T)[hidden_size, input_size]
        + R.special.next(V.Z, V.T)
    ) | [Transformation.TANH]
    template += R.h(V.T) <= (Transformation.REVERSE(R.h__z(V.T)) * R.h__n(V.T)) + (
        R.h__z(V.T) * R.h(V.Z)
    ) + R.special.next(V.Z, V.T)

    model = template.build(
        Settings(chain_pruning=False, iso_value_compression=False, optimizer=Adam(lr=0.001), error_function=MSE())
    )

    parameters = model.parameters()
    torch_parameters = [parameter.tolist() for parameter in rnn.parameters()]

    parameters["weights"][0] = [torch_parameters[0][i] for i in range(0, hidden_size)]
    parameters["weights"][2] = [torch_parameters[0][i] for i in range(1 * hidden_size, 1 * hidden_size + hidden_size)]
    parameters["weights"][5] = [torch_parameters[0][i] for i in range(2 * hidden_size, 2 * hidden_size + hidden_size)]

    parameters["weights"][1] = [torch_parameters[1][i] for i in range(0, hidden_size)]
    parameters["weights"][3] = [torch_parameters[1][i] for i in range(1 * hidden_size, 1 * hidden_size + hidden_size)]
    parameters["weights"][4] = [torch_parameters[1][i] for i in range(2 * hidden_size, 2 * hidden_size + hidden_size)]

    model.load_state_dict(parameters)

    dataset = Dataset(
        [
            Sample(
                R.h(sequence_len)[target.detach().numpy().tolist()],
                [
                    R.h0[[float(h) for h in h0[0]]],
                    *[R.f(i + 1)[[float(h) for h in torch_input[i]]] for i in range(sequence_len)],
                ],
            )
        ]
    )

    bd = model.build_dataset(dataset)

    optimizer = torch.optim.Adam(rnn.parameters(), lr=0.001)
    loss_fun = torch.nn.MSELoss()

    for _ in range(epochs):
        output, _ = rnn(torch_input, h0)
        loss = loss_fun(output[-1], target)

        optimizer.zero_grad(set_to_none=True)
        loss.backward()
        optimizer.step()

        result = model.train(bd, 1)
        assert np.allclose([float(x) for x in output[-1]], [float(x) for x in result[0][1]], atol=10e-5)



import csv
import io
from typing import List, Union, Callable

from neuralogic.core.constructs.relation import BaseRelation, WeightedRelation
from neuralogic.core.constructs.rule import Rule
from neuralogic.dataset.logic import Dataset
from neuralogic.dataset.csv import CSVDataset, CSVFile, Mode
from neuralogic.dataset.base import ConvertibleDataset

DatasetEntries = Union[BaseRelation, WeightedRelation, Rule]


class DBSource:
    __slots__ = (
        "relation_name",
        "table_name",
        "term_columns",
        "value_column",
        "default_value",
        "value_mapper",
        "skip_rows",
        "n_rows",
        "replace_empty_column",
        "sep",
    )

    def __init__(
        self,
        relation_name: str,
        table_name: str,
        term_columns: List[str],
        value_column: str | None = None,
        default_value: Union[float, int] = 1.0,
        value_mapper: Callable | None = None,
        skip_rows: int = 0,
        n_rows: int | None = None,
        replace_empty_column: Union[str, float, int] = 0,
        sep=",",
    ):
        self.table_name = table_name
        self.relation_name = relation_name
        self.sep = sep
        self.value_column = value_column
        self.default_value = default_value
        self.value_mapper = value_mapper
        self.term_columns = term_columns
        self.skip_rows = skip_rows
        self.n_rows = n_rows
        self.replace_empty_column = replace_empty_column

        if len(term_columns) == 0:
            raise NotImplementedError("Cannot create DBSource with zero terms")

    def to_csv(self, cursor) -> CSVFile:
        source = io.StringIO()

        columns = [term for term in self.term_columns]
        term_columns = list(range(len(columns)))
        value_column = None

        if self.value_column is not None:
            columns.append(self.value_column)
            value_column = len(columns) - 1

        if hasattr(cursor, "copy_to"):
            cursor.copy_to(source, self.table_name, sep=self.sep, null="", columns=columns)
        else:
            cursor.execute(f"SELECT {','.join(columns)} FROM {self.table_name}")
            results = cursor.fetchall()

            csv_writer = csv.writer(source, lineterminator="\n")
            csv_writer.writerows(results)

        source.seek(0)

        return CSVFile(
            self.relation_name,
            source,
            self.sep,
            value_column,
            self.default_value,
            self.value_mapper,
            term_columns,
            False,
            self.skip_rows,
            self.n_rows,
            self.replace_empty_column,
        )


class DBDataset(ConvertibleDataset):
    def __init__(
        self,
        connection,
        db_sources: Union[List[DBSource], DBSource],
        queries_db_source: DBSource | None = None,
        mode: Mode = Mode.ONE_EXAMPLE,
    ):
        self.connection = connection
        self.db_sources = [db_sources] if isinstance(db_sources, DBSource) else db_sources
        self.queries_db_source = queries_db_source
        self.mode = mode

    def add_db_source(self, db_source: DBSource):
        self.db_sources.append(db_source)

    def set_queries(self, db_source: DBSource):
        self.queries_db_source = db_source

    def to_dataset(self) -> Dataset:
        with self.connection.cursor() as cur:
            csv_files = [db_source.to_csv(cur) for db_source in self.db_sources]
            csv_queries = None if self.queries_db_source is None else self.queries_db_source.to_csv(cur)

        csv_dataset = CSVDataset(csv_files, csv_queries, self.mode)

        return csv_dataset.to_dataset()



import os
from typing import Optional, List

import jpype

from neuralogic.logging import _init_logging, LogHandler, add_log_handler


_is_initialized = False
_seed = int.from_bytes(os.urandom(4), byteorder="big")
_initial_seed = _seed
_rnd_generator = None
_max_memory_size = None
_graphviz_path = None

jvm_params = {
    "classpath": os.path.join(os.path.abspath(os.path.dirname(__file__)), "jar", "NeuraLogic.jar"),
}

jvm_options = ["-Xms1g"]


def set_max_memory_size(size: int):
    """
    Set maximum memory size that can be utilized by the backend (in gigabytes)

    Parameters
    ----------
    size : int
        The maximum memory size (in gigabytes)
    """
    global _max_memory_size
    _max_memory_size = size


def initial_seed() -> int:
    """
    Returns the initial/current random seed for a random number generator used in the backend.
    """
    return _initial_seed


def seed() -> int:
    """
    Sets the seed for a random number generator used in the backend to a random seed and returns the seed.
    """
    global _seed

    _seed = int.from_bytes(os.urandom(4), byteorder="big")

    if _rnd_generator is not None:
        _rnd_generator.setSeed(_seed)
    return _seed


def manual_seed(seed: int):
    """
    Sets the seed for a random number generator used in the backend to the passed ``seed``.

    Parameters
    ----------
    seed : int
        The seed for the random number generator.
    """
    global _seed

    _seed = seed

    if _rnd_generator is not None:
        _rnd_generator.setSeed(_seed)


def set_jvm_options(options: List[str]):
    """
    Set the jvm options - by default ``["-Xms1g"]``.

    Parameters
    ----------
    options : List[str]
        List of JVM options
    """
    global jvm_options
    jvm_options = options


def set_jvm_path(path: Optional[str]):
    """
    Set the JVM path.

    Parameters
    ----------
    path : Optional[str]
        The JVM path
    """
    global jvm_params

    if path is None:
        jvm_params.pop("jvmpath", None)
    else:
        jvm_params["jvmpath"] = path


def is_initialized() -> bool:
    """
    Check whether the NeuraLogic backend has been initialized
    """
    return _is_initialized


def set_graphviz_path(path: Optional[str]):
    """
    Set the default path to Graphviz

    Parameters
    ----------
    path : Optional[str]
        The Graphviz path
    """
    global _graphviz_path
    _graphviz_path = path


def get_default_graphviz_path() -> Optional[str]:
    """
    Get the default path to Graphviz
    """
    return _graphviz_path


def initialize(
    debug_mode: bool = False,
    debug_port: int = 12999,
    is_debug_server: bool = True,
    debug_suspend: bool = True,
    *,
    seed: Optional[int] = None,
    graphviz_path: Optional[str] = None,
    max_memory_size: Optional[int] = None,
    log_handler: Optional[LogHandler] = None,
    jar_path: Optional[str] = None,
    started_check: bool = True,
):
    """
    Initialize the NeuraLogic backend. This function is called implicitly when needed and should be called
    manually only for debugging.

    Parameters
    ----------
    debug_mode : bool
        Enable/Disable JVM debug mode.
    debug_port : int
        Port for the debugger to listen on. Default: ``12999``.
    is_debug_server : bool
        Act like server and listen for the debugger. Default: ``True``
    debug_suspend : bool
        Wait until the debugger is connected. Default: ``True``
    seed : Optional[int]
        The seed for the random number generator.
    graphviz_path : Optional[str]
        The Graphviz path
    max_memory_size : Optional[int]
        The maximum memory size (in gigabytes)
    log_handler: Optional[LogHandler]
        The handler for logging
    jar_path: Optional[str]
        The path to NeuraLogic java backend
    started_check: bool
        Whether to check if the NeuraLogic has been started. If true and it has been started - raise an Exception
    """
    global _is_initialized

    if _is_initialized and started_check:
        raise Exception("NeuraLogic already initialized")
    if _is_initialized:
        return

    if seed is not None:
        manual_seed(seed)

    if graphviz_path is not None:
        set_graphviz_path(graphviz_path)

    if max_memory_size is not None:
        set_max_memory_size(max_memory_size)

    if log_handler is not None:
        add_log_handler(log_handler)

    _is_initialized = True
    options = [*jvm_options]

    if _max_memory_size is not None:
        options.append(f"-Xmx{_max_memory_size}g")

    params = {**jvm_params}
    if jar_path is not None:
        params["classpath"] = jar_path

    if debug_mode:
        port = int(debug_port)
        server = "y" if is_debug_server else "n"
        suspend = "y" if debug_suspend else "n"

        debug_params = [
            "-Xint",
            "-Xdebug",
            "-Xnoagent",
            f"-Xrunjdwp:transport=dt_socket,server={server},address={port},suspend={suspend}",
        ]

        jpype.startJVM(*options, *debug_params, **params)
    else:
        jpype.startJVM(*options, **params)
    _init_logging()



from neuralogic.core.constructs.function.function import AggregationFunction
from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Transformation, Aggregation, Combination
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.general.mlp import MLP
from neuralogic.nn.module.module import Module


class GENConv(Module):
    r"""
    GENConv layer from `"DeeperGCN: All You Need to Train Deeper GCNs" <https://arxiv.org/abs/2006.07739>`_.

    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    edge_name : str
        Edge predicate name to use for neighborhood relations.
    aggregation : AggregationFunction
        The aggregation function.
        Default: ``Aggregation.SOFTMAX``
    num_layers : int
        The number of MLP layers.
        Default: ``2``
    expansion : int
        The expansion factor of hidden channels in MLP.
        Default: ``2``
    eps : float
        :math:`\epsilon`-value.
        Default: ``0.0``
    train_eps : bool
        Is ``eps`` trainable parameter.
        Default: ``false``
    edge_dim : int | None
        Dimension of edge features (``None`` is projection to ``in_channels`` is not needed).
        Default: ``None``
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        edge_name: str,
        aggregation: AggregationFunction = Aggregation.SOFTMAX,
        num_layers: int = 2,
        expansion: int = 2,
        eps: float = 1e-7,
        train_eps: bool = False,
        edge_dim: int | None = None,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.edge_name = edge_name

        self.aggregation = aggregation
        self.num_layers = num_layers
        self.expansion = expansion

        self.eps = eps
        self.train_eps = train_eps

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.edge_dim: int | None = edge_dim

    def __call__(self):
        feat_sum = R.get(f"{self.output_name}__gen_feat_sum")
        feat_agg = R.get(f"{self.output_name}__gen_feat_agg")
        out = R.get(f"{self.output_name}__gen_out")
        x = R.get(self.feature_name)
        e = R.get(self.edge_name)

        eps = R.get(f"{self.output_name}__eps")
        v_eps = eps[self.eps]
        if not self.train_eps:
            v_eps = v_eps.fixed()

        e_proj = []
        if self.edge_dim is not None and self.out_channels != self.edge_dim:
            e = R.get(f"{self.output_name}__gen_edge_proj")
            e_proj = [e(V.I, V.J)[self.out_channels, self.edge_dim] <= R.get(self.edge_name)(V.I, V.J)]

        channels = [self.out_channels]
        for _ in range(self.num_layers - 1):
            channels.append(self.out_channels * self.expansion)
        channels.append(self.out_channels)

        mlp = MLP(channels, self.output_name, f"{self.output_name}__gen_out", Transformation.IDENTITY)

        j_feat = x(V.J)
        i_feat = x(V.I)
        if self.in_channels != self.out_channels:
            j_feat = x(V.J)[self.out_channels, self.in_channels]
            i_feat = x(V.I)[self.out_channels, self.in_channels]

        return [
            v_eps,
            *e_proj,
            (feat_sum(V.I, V.J) <= (j_feat, e(V.J, V.I)))
            | Metadata(transformation=Transformation.RELU, combination=Combination.SUM),
            (feat_agg(V.I) <= (feat_sum(V.I, V.J), eps))
            | Metadata(aggregation=self.aggregation, combination=Combination.SUM),
            (out(V.I) <= (i_feat, feat_agg(V.I))) | Metadata(combination=Combination.SUM),
            *mlp(),
        ]



import math

from neuralogic.core.constructs.function import Transformation, Combination, Aggregation
from neuralogic.core.constructs.factories import R
from neuralogic.nn.module.module import Module


class Attention(Module):
    r"""
    A single-head attention module based on `"Attention Is All You Need" <https://arxiv.org/abs/1706.03762>`_.

    Parameters
    ----------

    embed_dim : int
        The number of expected features.
    output_name : str
        Output (head) predicate name of the module.
    query_name : str
        The name of the queries predicate.
    key_name : str
        The name of the keys predicate.
    value_name : str
        The name of the values predicate.
    mask_name : str, optional
        The name of the input mask predicate. Default: ``None``
    arity : int
        Arity of the input and output predicates. Default: ``1``
    """

    def __init__(
        self,
        embed_dim: int,
        output_name: str,
        query_name: str,
        key_name: str,
        value_name: str,
        mask_name: str | None = None,
        arity: int = 1,
    ):
        self.embed_dim = embed_dim
        self.output_name = output_name
        self.query_name = query_name
        self.key_name = key_name
        self.value_name = value_name
        self.mask_name = mask_name
        self.arity = arity

    def __call__(self):
        terms = [f"X{i}" for i in range(self.arity - 1)]

        k_terms = [*terms, "Y"]
        h_terms = [*terms, "X", "Y"]
        q_terms = [*terms, "X"]

        d_k = 1 / math.sqrt(self.embed_dim)

        dk_rel = R.get(f"{self.output_name}__dk")
        dot_rel = R.get(f"{self.output_name}__dot")

        metadata = [Combination.PRODUCT, Aggregation.SOFTMAX(agg_terms=["Y"])]
        out_metadata = [Combination.PRODUCT, Aggregation.SUM]

        attention_product_rules = [
            (dot_rel(h_terms) <= (dk_rel, R.get(self.key_name)(k_terms).T, R.get(self.query_name)(q_terms))) | metadata,
        ]

        if self.mask_name is not None:
            attention_product_rules[0].body.append(R.hidden.get(self.mask_name)(h_terms))

        return [
            dk_rel[d_k].fixed(),
            *attention_product_rules,
            (R.get(self.output_name)(q_terms) <= (dot_rel(h_terms), R.get(self.value_name)(k_terms))) | out_metadata,
        ]


class MultiheadAttention(Module):
    r"""
    A multi-head attention module based on `"Attention Is All You Need" <https://arxiv.org/abs/1706.03762>`_.

    Parameters
    ----------

    embed_dim : int
        The number of expected features.
    num_heads : int
        The number of heads.
    output_name : str
        Output (head) predicate name of the module.
    query_name : str
        The name of the queries predicate.
    key_name : str
        The name of the keys predicate.
    value_name : str
        The name of the values predicate.
    vdim : Optional[int]
        Total number of features for values.
    kdim : Optional[int]
        Total number of features for keys.
    mask_name : str, optional
        The name of the input mask predicate. Default: ``None``
    arity : int
        Arity of the input and output predicates. Default: ``1``
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        output_name: str,
        query_name: str,
        key_name: str,
        value_name: str,
        vdim: int | None = None,
        kdim: int | None = None,
        mask_name: str | None = None,
        arity: int = 1,
    ):
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.output_name = output_name
        self.queries = query_name
        self.keys = key_name
        self.values = value_name
        self.vdim: int = vdim if vdim is not None else embed_dim
        self.kdim: int = kdim if kdim is not None else embed_dim
        self.mask_name = mask_name
        self.arity = arity

    def __call__(self):
        terms = [f"X{i}" for i in range(self.arity)]
        dim = self.embed_dim

        q_weight = f"{self.output_name}_qw"
        k_weight = f"{self.output_name}_kw"
        v_weight = f"{self.output_name}_vw"

        q_proj_name = f"{self.output_name}__qproj"
        k_proj_name = f"{self.output_name}__kproj"
        v_proj_name = f"{self.output_name}__vproj"

        q_proj = R.get(q_proj_name)
        k_proj = R.get(k_proj_name)
        v_proj = R.get(v_proj_name)
        output_rel = R.get(self.output_name)

        attention_name = f"{self.output_name}__attention"

        attention = Attention(
            dim // self.num_heads, attention_name, q_proj_name, k_proj_name, v_proj_name, self.mask_name, self.arity
        )

        if self.num_heads != 1:
            size = self.embed_dim / self.num_heads
            attention.arity += 1

            attention_concat = []
            multihead_rules = []

            for i in range(self.num_heads):
                meta = [Transformation.SLICE(rows=(i * size, (i + 1) * size))]
                multihead_rules.append((q_proj(i, *terms) <= R.get(self.queries)(terms)[q_weight:dim, dim]) | meta)
                multihead_rules.append((v_proj(i, *terms) <= R.get(self.values)(terms)[v_weight:dim, self.vdim]) | meta)
                multihead_rules.append((k_proj(i, *terms) <= R.get(self.keys)(terms)[k_weight:dim, self.kdim]) | meta)
                attention_concat.append(R.get(attention_name)(i, *terms))

            multihead_rules.append((output_rel(terms)[dim, dim] <= attention_concat) | [Combination.CONCAT])
        else:
            multihead_rules = [
                q_proj(terms)[q_weight:dim, dim] <= R.get(self.queries)(terms),
                v_proj(terms)[v_weight:dim, self.vdim] <= R.get(self.values)(terms),
                k_proj(terms)[k_weight:dim, self.kdim] <= R.get(self.keys)(terms),
                output_rel(terms)[dim, dim] <= R.get(attention_name)(terms),
            ]

        return [*attention(), *multihead_rules]



import pytest

import torch
import numpy as np

from neuralogic.core import Template, R, Settings
from neuralogic.core import F
from neuralogic.dataset import Dataset, Sample


@pytest.mark.parametrize(
    "torch_fun, fun",
    [
        (torch.sigmoid, F.sigmoid),
        (torch.tanh, F.tanh),
        (torch.sign, F.signum),
        (torch.relu, F.relu),
        (torch.exp, F.exp),
        (torch.log, F.log),
    ],
)
def test_transformation_body_function(torch_fun, fun):
    data = torch.rand((2, 3))

    torch_result = torch_fun(data).detach().numpy().round(3)

    template = Template()
    template += R.h <= fun(R.input)

    model = template.build(Settings(iso_value_compression=False, chain_pruning=False))
    dataset = Dataset([Sample(R.h, R.input[data.tolist()])])

    built_dataset = model.build_dataset(dataset)
    results = np.array(model.test(built_dataset)).round(3)

    assert np.allclose(torch_result, results, atol=0.0001)


def test_slice_function():
    data = np.array(
        [
            [1, 2, 3, 4],
            [5, 6, 7, 8],
            [9, 10, 11, 12],
            [13, 14, 15, 16],
        ]
    )

    res = np.array(
        [
            [5, 6, 7, 8],
            [9, 10, 11, 12],
        ]
    )

    template = Template()
    template += R.h <= F.slice(R.input, rows=(1, 3))

    model = template.build(Settings(iso_value_compression=False, chain_pruning=False))
    dataset = Dataset([Sample(R.h, [R.input[data]])])

    built_dataset = model.build_dataset(dataset)
    results = np.array(model.test(built_dataset))

    assert np.allclose(res, results)

    template = Template()
    template += (R.h <= R.input) | [F.slice(rows=(1, 3))]

    model = template.build(Settings(iso_value_compression=False, chain_pruning=False))
    dataset = Dataset(Sample(R.h, [R.input[data]]))

    built_dataset = model.build_dataset(dataset)
    results = np.array(model.test(built_dataset))

    assert np.allclose(res, results)

    template = Template()
    template += R.h <= R.input
    template += R.h / 0 | [F.slice(rows=(1, 3))]

    model = template.build(Settings(iso_value_compression=False, chain_pruning=False))
    dataset = Dataset(Sample(R.h, [R.input[data]]))

    built_dataset = model.build_dataset(dataset)
    results = np.array(model.test(built_dataset))

    assert np.allclose(res, results)



import jpype

from neuralogic.core.builder import Builder
from neuralogic.core.builder.components import NeuralSample, Grounding


class BuiltDataset:
    """BuiltDataset represents an already built dataset - that is, a dataset that has been grounded and neuralized."""

    __slots__ = "_samples", "_batch_size"

    def __init__(self, samples: list[NeuralSample], batch_size: int):
        self._samples = samples
        self._batch_size = batch_size

    def __len__(self):
        return len(self._samples)

    def __getitem__(self, item):
        return self._samples[item]

    def __iter__(self):
        return iter(self._samples)


class GroundedDataset:
    """GroundedDataset represents grounded examples that are not neuralized yet."""

    __slots__ = "_groundings", "_groundings_list", "_builder"

    def __init__(self, groundings, builder: Builder):
        self._builder = builder
        self._groundings = groundings
        self._groundings_list = [Grounding(g) for g in self._groundings]

    def __getitem__(self, item) -> Grounding:
        return self._groundings_list[item]

    def __len__(self) -> int:
        return len(self._groundings_list)

    def __iter__(self):
        return iter(self._groundings_list)

    def neuralize(self, *, batch_size: int = 1, progress: bool = False) -> BuiltDataset:
        return BuiltDataset(self._builder.neuralize(self._groundings.stream(), progress, len(self)), batch_size)



import torch


class NeuralogicOptTensor(torch.Tensor):
    _neuralogic_weight_updater = None
    _neuralogic_weight = None
    _neuralogic_value_factory = None

    def add_(self, other, *args, **kwargs):
        r = super().add_(other, *args, **kwargs)
        if self._neuralogic_weight is None:
            return r

        self.__sync_neuralogic__()

        return r

    def addcdiv_(self, tensor1, tensor2, *args, **kwargs):
        r = super().addcdiv_(tensor1, tensor2, *args, **kwargs)
        if self._neuralogic_weight is None:
            return r

        self.__sync_neuralogic__()

        return r

    def addcmul_(self, tensor1, tensor2, *args, **kwargs):
        r = super().addcmul_(tensor1, tensor2, *args, **kwargs)
        if self._neuralogic_weight is None:
            return r

        self.__sync_neuralogic__()

        return r

    @property
    def grad(self):
        grad_list = self._neuralogic_weight_updater.weightUpdates
        index = self._neuralogic_weight.index

        if len(grad_list) <= index:
            return None

        value = grad_list[index]
        if value is None:
            return None

        p = -torch.tensor(self._neuralogic_value_factory.from_java(value))
        return p

    @staticmethod
    def create(weight, data, weight_updater, value_factory):
        tensor = torch.tensor(data, requires_grad=True)
        tensor.__class__ = NeuralogicOptTensor
        tensor._neuralogic_weight = weight
        tensor._neuralogic_weight_updater = weight_updater
        tensor._neuralogic_value_factory = value_factory

        return tensor

    def __sync_neuralogic__(self):
        size = self.size()
        weight_value = self._neuralogic_weight.value

        if len(size) == 0:
            weight_value.set(0, float(self))
            return

        if len(size) == 1:
            for i, val in enumerate(self):
                weight_value.set(i, float(val))
            return

        cols = size[0]

        for i, values in enumerate(self):
            for j, val in enumerate(values):
                weight_value.set(i * cols + j, float(val))



from typing import Any
import weakref

from neuralogic.core.enums import Grounder
from neuralogic.nn.init import Initializer, Uniform
from neuralogic.nn.loss import MSE, ErrorFunction
from neuralogic.core.settings.settings_proxy import SettingsProxy
from neuralogic.optim import Optimizer, Adam


class Settings:
    def __init__(
        self,
        *,
        optimizer: Optimizer = Adam(),
        epochs: int = 3000,
        error_function: ErrorFunction = MSE(),
        initializer: Initializer = Uniform(),
        iso_value_compression: bool = True,
        chain_pruning: bool = True,
        prune_only_identities: bool = False,
        grounder: Grounder = Grounder.BUP,
    ):
        self.params = locals().copy()
        self.params.pop("self")
        self._proxies: weakref.WeakSet[SettingsProxy] = weakref.WeakSet()

        self.kw_params: dict[str, Any] = {}

    @property
    def iso_value_compression(self) -> bool:
        return self.params["iso_value_compression"]

    @iso_value_compression.setter
    def iso_value_compression(self, iso_value_compression: bool):
        self._update("iso_value_compression", iso_value_compression)

    @property
    def chain_pruning(self) -> bool:
        return self.params["chain_pruning"]

    @chain_pruning.setter
    def chain_pruning(self, chain_pruning: bool):
        self._update("chain_pruning", chain_pruning)

    @property
    def prune_only_identities(self) -> bool:
        return self.params["prune_only_identities"]

    @prune_only_identities.setter
    def prune_only_identities(self, prune_only_identities: bool):
        self._update("prune_only_identities", prune_only_identities)

    @property
    def grounder(self) -> Grounder:
        return self.params["grounder"]

    @grounder.setter
    def grounder(self, grounder: Grounder):
        self._update("grounder", grounder)

    @property
    def optimizer(self) -> Optimizer:
        return self.params["optimizer"]

    @optimizer.setter
    def optimizer(self, optimizer: Optimizer):
        self._update("optimizer", optimizer)

    @property
    def error_function(self) -> ErrorFunction:
        return self.params["error_function"]

    @error_function.setter
    def error_function(self, error_function: ErrorFunction):
        self._update("error_function", error_function)

    @property
    def epochs(self) -> int:
        return self.params["epochs"]

    @epochs.setter
    def epochs(self, epochs: int):
        self._update("epochs", epochs)

    @property
    def initializer(self) -> Initializer:
        return self.params["initializer"]

    @initializer.setter
    def initializer(self, initializer: Initializer):
        self._update("initializer", initializer)

    def create_proxy(self) -> SettingsProxy:
        proxy = SettingsProxy(**self.params)
        self._proxies.add(proxy)

        for k, v in self.kw_params.items():
            proxy[k] = v

        return proxy

    def create_disconnected_proxy(self) -> SettingsProxy:
        proxy = SettingsProxy(**self.params)
        for k, v in self.kw_params.items():
            proxy[k] = v
        return proxy

    def __setitem__(self, key, value):
        for proxy in self._proxies.copy():
            proxy[key] = value
        self.kw_params[key] = value

    def __getitem__(self, item):
        return self.kw_params[item]

    def _update(self, parameter: str, value: Any) -> None:
        if parameter not in self.params:
            raise NotImplementedError
        self.params[parameter] = value

        for proxy in self._proxies.copy():
            proxy.__setattr__(parameter, value)



from typing import List, Collection

import torch

from neuralogic.core.constructs.java_objects import ValueFactory
from neuralogic.core.torch.network_output import PyNeuraLogicNetworkOutput
from neuralogic.core.torch.tensor import NeuralogicOptTensor


class TorchNeuralModule:
    def tensor_parameters(
        self, tensor_parameters, weight_updater, value_factory, neural_model
    ) -> List[NeuralogicOptTensor]:
        if tensor_parameters is None:
            tensor_parameters = [
                NeuralogicOptTensor.create(
                    weight,
                    ValueFactory.from_java(weight.value),
                    weight_updater,
                    value_factory,
                )
                for weight in neural_model.getAllWeights()
                if weight.isLearnable
            ]

        return tensor_parameters

    def update_tensor_parameters(self, tensor_parameters):
        if tensor_parameters is None:
            return

        for param in tensor_parameters:
            param.data = torch.tensor(ValueFactory.from_java(param._neuralogic_weight.value))

    def forward(self, model, samples, results) -> PyNeuraLogicNetworkOutput:
        if not isinstance(samples, Collection):
            return PyNeuraLogicNetworkOutput.apply(
                samples, model, torch.tensor(results[0], dtype=torch.float, requires_grad=True)
            )
        return PyNeuraLogicNetworkOutput.apply(
            samples, model, torch.tensor(results, dtype=torch.float, requires_grad=True)
        )



from neuralogic.core.constructs.function.enum import Transformation, Combination, Aggregation, F
from neuralogic.core.constructs.function.function import Function
from neuralogic.core.constructs.function.function_container import FContainer

__all__ = ["Transformation", "Combination", "Aggregation", "F", "Function", "FContainer"]



from typing import Any, Collection

import torch
from torch.autograd import Function


class PyNeuraLogicNetworkOutput(Function):
    @staticmethod
    def forward(ctx, samples, model, tensor):
        ctx.model = model
        ctx.samples = samples
        ctx.shape = tensor.shape

        return tensor

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        model = ctx.model
        samples = ctx.samples

        if not isinstance(samples, Collection):
            model._backprop(samples, -grad_outputs[0])
        else:
            for sample, grad in zip(samples, grad_outputs[0]):
                model._backprop(sample, -grad)
        return None, None, torch.zeros(ctx.shape, dtype=torch.float64)



from typing import Dict, Any

from neuralogic.core import Settings, Transformation
from neuralogic.core.constructs.function import Function
from neuralogic.nn.init import Initializer, Uniform
from neuralogic.nn.loss import SoftEntropy, ErrorFunction
from neuralogic.optim import SGD, Optimizer

import pytest


@pytest.mark.parametrize(
    "parameters",
    [
        {
            "optimizer": SGD(0.5),
            "epochs": 100,
            "error_function": SoftEntropy(),
            "initializer": Uniform(5.0),
            "initializer_uniform_scale": 5.0,
        }
    ],
)
def test_settings_proxy_properties_setting(parameters: Dict[str, Any]) -> None:
    """Tests propagation of changes on settings to its proxies"""
    settings = Settings()
    settings_proxy = settings.create_proxy()

    for key, value in parameters.items():
        if key == "initializer_uniform_scale":
            assert settings.initializer.scale == settings_proxy.__getattribute__(key)
            continue

        if isinstance(value, (int, float)):
            assert settings.__getattribute__(key) == settings_proxy.__getattribute__(key)
        elif isinstance(settings.__getattribute__(key), (ErrorFunction, Initializer, Function, Optimizer)):
            assert str(settings.__getattribute__(key)) == str(settings_proxy.__getattribute__(key))
        else:
            assert settings.__getattribute__(key) == str(settings_proxy.__getattribute__(key))

    for key, value in parameters.items():
        settings.__setattr__(key, value)

    for key, value in parameters.items():
        if key == "initializer_uniform_scale":
            assert settings.initializer.scale == settings_proxy.__getattribute__(key)
            continue

        if isinstance(value, (int, float)):
            assert settings.__getattribute__(key) == settings_proxy.__getattribute__(key)
        elif isinstance(settings.__getattribute__(key), (ErrorFunction, Initializer, Function, Optimizer)):
            assert str(settings.__getattribute__(key)) == str(settings_proxy.__getattribute__(key))
        else:
            assert settings.__getattribute__(key) == str(settings_proxy.__getattribute__(key))


def test_settings_proxies_creation() -> None:
    """Tests settings proxies creation and dereferencing"""
    settings = Settings()

    settings_proxy = settings.create_proxy()  # We have one proxy
    assert len(settings._proxies) == 1

    second_proxy = settings.create_proxy()  # Now we have two proxies
    assert len(settings._proxies) == 2

    # We are creating the third one, but the original `settings_proxy` is dereferenced
    settings_proxy = settings.create_proxy()
    assert len(settings._proxies) == 2

    del second_proxy
    assert len(settings._proxies) == 1

    del settings_proxy
    assert len(settings._proxies) == 0



from neuralogic.logging import LogHandler, add_log_handler
from neuralogic.core import R, V, C, F, Template, Settings, Relation, Var, Const
from neuralogic.setup import (
    seed,
    set_max_memory_size,
    manual_seed,
    initial_seed,
    initialize,
    is_initialized,
    set_graphviz_path,
    set_jvm_path,
    set_jvm_options,
    get_default_graphviz_path,
)


__all__ = [
    "R",
    "V",
    "C",
    "F",
    "Relation",
    "Var",
    "Const",
    "Template",
    "Settings",
    "LogHandler",
    "add_log_handler",
    "set_max_memory_size",
    "manual_seed",
    "initial_seed",
    "seed",
    "initialize",
    "is_initialized",
    "set_graphviz_path",
    "set_jvm_path",
    "set_jvm_options",
    "get_default_graphviz_path",
]



from typing import Iterable, Union

import numpy as np

from neuralogic.core.constructs.function.enum import Combination, Transformation
from neuralogic.core.constructs.function import FContainer
from neuralogic.core.constructs.predicate import Predicate
from neuralogic.core.constructs import rule, factories
from neuralogic.core.constructs.function.function import TransformationFunction, CombinationFunction


class BaseRelation:
    __slots__ = "predicate", "function", "terms", "negated"

    def __init__(
        self,
        predicate: Predicate,
        terms=None,
        function: Union[TransformationFunction, CombinationFunction, None] = None,
        negated: bool = False,
    ):
        self.predicate = predicate
        self.function = function
        self.negated = negated
        self.terms = []

        if not isinstance(terms, Iterable) or isinstance(terms, str):
            terms = [terms]

        for term in terms:
            if term is None:
                continue

            if isinstance(term, list):
                self.terms.extend(term)
            else:
                self.terms.append(term)

    def __neg__(self) -> "BaseRelation":
        return self.attach_activation_function(Transformation.REVERSE)

    def __invert__(self) -> "BaseRelation":
        if self.function is not None:
            raise ValueError(f"Cannot negate relation {self} with attached function.")

        predicate = Predicate(self.predicate.name, self.predicate.arity, True, self.predicate.special)
        relation = BaseRelation(predicate, self.terms, self.function, not self.negated)

        return relation

    @property
    def T(self) -> "BaseRelation":
        return self.attach_activation_function(Transformation.TRANSP)

    def attach_activation_function(self, function: Union[TransformationFunction, CombinationFunction]):
        if self.negated:
            raise ValueError(f"Cannot attach function to negated relation {self}")
        relation = self.__copy__()
        relation.function = function
        return relation

    def __truediv__(self, other):
        if not isinstance(other, int) or self.predicate.arity != 0 or other < 0:
            raise NotImplementedError

        name, hidden, special = self.predicate.name, self.predicate.hidden, self.predicate.special
        return factories.AtomFactory.Predicate.get_predicate(name, other, hidden, special)

    def __call__(self, *args) -> "BaseRelation":
        if self.terms:
            raise Exception

        if len(args) == 1 and isinstance(args[0], Iterable) and not isinstance(args[0], str):
            terms = list(args[0])
        else:
            terms = list(args)
        arity = len(terms)

        name, hidden, special = self.predicate.name, self.predicate.hidden, self.predicate.special
        predicate = factories.AtomFactory.Predicate.get_predicate(name, arity, hidden, special)

        return BaseRelation(predicate, terms, self.function, self.negated)

    def __getitem__(self, item) -> "WeightedRelation":
        if self.predicate.hidden or self.predicate.special:
            raise ValueError(f"Special/Hidden relation {self} cannot have learnable parameters.")
        return WeightedRelation(item, self.predicate, False, self.terms, self.function)

    def __le__(self, other: Union[Iterable["BaseRelation"], "BaseRelation"]) -> rule.Rule:
        return rule.Rule(self, other)

    def to_str(self, end=False) -> str:
        end = "." if end else ""

        if self.terms:
            terms = ", ".join([str(term) for term in self.terms])

            if self.negated:
                return f"!{self.predicate.to_str()}({terms}){end}"
            if self.function:
                literal = f"{self.predicate.to_str()}({terms})"
                return f"{self.function.wrap(literal)}{end}"
            return f"{self.predicate.to_str()}({terms}){end}"

        if self.negated:
            return f"!{self.predicate.to_str()}{end}"
        if self.function:
            return f"{self.function.wrap(self.predicate.to_str())}{end}"
        return f"{self.predicate.to_str()}{end}"

    def __str__(self) -> str:
        return self.to_str(True)

    def __repr__(self) -> str:
        return self.__str__()

    def __copy__(self):
        relation = BaseRelation.__new__(BaseRelation)
        relation.function = self.function
        relation.terms = self.terms
        relation.predicate = self.predicate
        relation.negated = self.negated

        return relation

    def __and__(self, other) -> rule.RuleBody:
        if isinstance(other, BaseRelation):
            return rule.RuleBody(self, other)
        raise NotImplementedError

    def __add__(self, other):
        return FContainer((self, other), Combination.SUM)

    def __mul__(self, other):
        return FContainer((self, other), Combination.ELPRODUCT)

    def __matmul__(self, other):
        return FContainer((self, other), Combination.PRODUCT)


class WeightedRelation(BaseRelation):
    __slots__ = "weight", "weight_name", "is_fixed"

    def __init__(
        self,
        weight,
        predicate: Predicate,
        fixed=False,
        terms=None,
        function: Union[TransformationFunction, CombinationFunction, None] = None,
    ):
        super().__init__(predicate, terms, function, False)

        self.weight = weight
        self.weight_name = None
        self.is_fixed = fixed

        if isinstance(weight, slice):
            self.weight_name = str(weight.start)
            self.weight = weight.stop
        elif isinstance(weight, tuple) and isinstance(weight[0], slice):
            self.weight_name = str(weight[0].start)
            self.weight = (weight[0].stop, *weight[1:])

        if isinstance(weight, np.ndarray):
            self.weight = weight.tolist()

    def fixed(self) -> "WeightedRelation":
        if self.is_fixed:
            raise Exception(f"Weighted relation {self} is already fixed")
        return WeightedRelation(self.weight, self.predicate, True, self.terms, self.function)

    def to_str(self, end=False):
        if isinstance(self.weight, tuple):
            weight = f"{{{', '.join(str(w) for w in self.weight)}}}"
        else:
            weight = str(self.weight)
        if self.weight_name:
            weight = f"${self.weight_name}={weight}"

        if self.is_fixed:
            return f"<{weight}> {super().to_str(end)}"
        return f"{weight} {super().to_str(end)}"

    def __str__(self) -> str:
        return self.to_str(True)

    def __repr__(self) -> str:
        return self.__str__()

    def __call__(self, *args) -> BaseRelation:
        raise NotImplementedError(f"Cannot assign terms to weighted relation {self.predicate}")

    def __getitem__(self, item) -> "WeightedRelation":
        raise NotImplementedError(f"Cannot assign weight to weighted relation {self.predicate}")

    def attach_activation_function(self, function: Union[Transformation, Combination]):
        raise NotImplementedError(
            f"Cannot attach a function to weighted relation {self}. Attach the function before adding weights."
        )

    @property
    def T(self) -> "WeightedRelation":
        raise NotImplementedError(
            f"Cannot transpose weighted relation {self} Apply the transposition before adding weights."
        )

    def __invert__(self) -> "WeightedRelation":
        raise NotImplementedError(f"Weighted relations ({self}) cannot be negated.")

    def __neg__(self) -> "WeightedRelation":
        raise NotImplementedError(
            f"Cannot negate weighted relation {self} Apply the reverse function before adding weights."
        )

    def __copy__(self):
        relation = WeightedRelation.__new__(WeightedRelation)

        relation.predicate = self.predicate
        relation.function = self.function
        relation.terms = self.terms
        relation.weight = self.weight
        relation.is_fixed = self.is_fixed
        relation.negated = self.negated

        return relation



from neuralogic.core.constructs.function import Transformation
from neuralogic.core.constructs.factories import R
from neuralogic.nn.module.module import Module
from neuralogic.nn.module.general.mlp import MLP
from neuralogic.nn.module.general.attention import MultiheadAttention


class Transformer(Module):
    r"""
    A transformer module based on `"Attention Is All You Need" <https://arxiv.org/abs/1706.03762>`_.

    Parameters
    ----------

    input_dim : int
        The number of expected features.
    num_heads : int
        The number of heads in the multi-head attention module.
    dim_feedforward : int
        The dimension of the feedforward network.
    output_name : str
        Output (head) predicate name of the module.
    src_name : str
        The name of the predicate of the input to the encoder.
    tgt_name : str
        The name of the predicate of the input to the decoder.
    src_mask_name : str, optional
        The name of the predicate of the encoder input mask. Default: ``None``
    tgt_mask_name : str, optional
        The name of the predicate of the decoder input mask. Default: ``None``
    memory_mask_name : str, optional
        The name of the predicate of the encoder output mask. Default: ``None``
    arity : int
        Arity of the input and output predicate. Default: ``1``
    """

    def __init__(
        self,
        input_dim: int,
        num_heads: int,
        dim_feedforward: int,
        output_name: str,
        src_name: str,
        tgt_name: str,
        src_mask_name: str | None = None,
        tgt_mask_name: str | None = None,
        memory_mask_name: str | None = None,
        arity: int = 1,
    ):
        self.input_dim = input_dim
        self.num_heads = num_heads
        self.dim_feedforward = dim_feedforward
        self.output_name = output_name
        self.src_name = src_name
        self.tgt_name = tgt_name
        self.src_mask_name = src_mask_name
        self.tgt_mask_name = tgt_mask_name
        self.memory_mask_name = memory_mask_name
        self.arity = arity

    def __call__(self):
        encoder = TransformerEncoder(
            self.input_dim,
            self.num_heads,
            self.dim_feedforward,
            f"{self.output_name}__encoder",
            self.src_name,
            self.src_mask_name,
            self.arity,
        )

        decoder = TransformerDecoder(
            self.input_dim,
            self.num_heads,
            self.dim_feedforward,
            self.output_name,
            self.tgt_name,
            f"{self.output_name}__encoder",
            self.tgt_mask_name,
            self.memory_mask_name,
            self.arity,
        )

        return [
            *encoder(),
            *decoder(),
        ]


class EncoderBlock(Module):
    def __init__(
        self,
        input_dim: int,
        num_heads: int,
        dim_feedforward: int,
        output_name: str,
        query_name: str,
        key_name: str,
        value_name: str,
        mask_name: str | None = None,
        arity: int = 1,
        mlp: bool = True,
    ):
        self.input_dim = input_dim
        self.num_heads = num_heads
        self.dim_feedforward = dim_feedforward
        self.output_name = output_name
        self.query_name = query_name
        self.key_name = key_name
        self.value_name = value_name
        self.mask_name = mask_name
        self.arity = arity
        self.mlp = mlp

    def __call__(self):
        terms = [f"X{i}" for i in range(self.arity)]

        attn_name = f"{self.output_name}__mhattn"
        norm_name = f"{self.output_name}__norm"
        mlp_name = f"{self.output_name}__mlp"

        output_rel = R.get(self.output_name)
        dim = self.input_dim
        data_name = self.query_name

        attention = MultiheadAttention(
            dim,
            self.num_heads,
            attn_name,
            self.query_name,
            self.key_name,
            self.value_name,
            mask_name=self.mask_name,
            arity=self.arity,
        )

        if self.mlp:
            dims = [dim, self.dim_feedforward, dim]
            mlp = MLP(dims, mlp_name, norm_name, [Transformation.RELU, Transformation.IDENTITY], self.arity)

            return [
                *attention(),
                (R.get(norm_name)(terms) <= (R.get(attn_name)(terms), R.get(data_name)(terms))) | [Transformation.NORM],
                *mlp(),
                (output_rel(terms) <= (R.get(norm_name)(terms), R.get(mlp_name)(terms))) | [Transformation.NORM],
            ]

        return [
            *attention(),
            (output_rel(terms) <= (R.get(attn_name)(terms), R.get(data_name)(terms))) | [Transformation.NORM],
        ]


class TransformerEncoder(EncoderBlock):
    r"""
    A transformer encoder module based on `"Attention Is All You Need" <https://arxiv.org/abs/1706.03762>`_.

    Parameters
    ----------

    input_dim : int
        The number of expected features.
    num_heads : int
        The number of heads in the multi-head attention module.
    dim_feedforward : int
        The dimension of the feedforward network.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        The name of the predicate of the input sequence.
    mask_name : str, optional
        The name of the predicate of the input sequence mask. Default: ``None``
    arity : int
        Arity of the input and output predicate. Default: ``1``
    """

    def __init__(
        self,
        input_dim: int,
        num_heads: int,
        dim_feedforward: int,
        output_name: str,
        input_name: str,
        mask_name: str | None = None,
        arity: int = 1,
    ):
        super().__init__(
            input_dim,
            num_heads,
            dim_feedforward,
            output_name,
            input_name,
            input_name,
            input_name,
            mask_name,
            arity,
            True,
        )


class TransformerDecoder(Module):
    r"""
    A transformer decoder module based on `"Attention Is All You Need" <https://arxiv.org/abs/1706.03762>`_.

    Parameters
    ----------

    input_dim : int
        The number of expected features.
    num_heads : int
        The number of heads in the multi-head attention module.
    dim_feedforward : int
        The dimension of the feedforward network.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        The name of the predicate of the input sequence.
    input_name : str
        The name of the input encoder.
    mask_name : str, optional
        The name of the predicate of the decoder input sequence mask. Default: ``None``
    memory_mask_name : str, optional
        The name of the predicate of the encoder output mask. Default: ``None``
    arity : int
        Arity of the input and output predicate. Default: ``1``
    """

    def __init__(
        self,
        input_dim: int,
        num_heads: int,
        dim_feedforward: int,
        output_name: str,
        input_name: str,
        encoder_name: str,
        mask_name: str | None = None,
        memory_mask_name: str | None = None,
        arity: int = 1,
    ):
        self.input_dim = input_dim
        self.num_heads = num_heads
        self.dim_feedforward = dim_feedforward
        self.output_name = output_name
        self.input_name = input_name
        self.encoder_name = encoder_name
        self.mask_name = mask_name
        self.memory_mask_name = memory_mask_name
        self.arity = arity

    def __call__(self):
        data_name = self.input_name
        dim = self.input_dim

        tmp_encoder_out = f"{self.output_name}__decoder"
        encoder_name = self.encoder_name
        mlp_dim = self.dim_feedforward

        enc_block_one = EncoderBlock(
            dim,
            self.num_heads,
            mlp_dim,
            tmp_encoder_out,
            data_name,
            data_name,
            data_name,
            self.mask_name,
            self.arity,
            False,
        )

        enc_block_two = EncoderBlock(
            dim,
            self.num_heads,
            mlp_dim,
            self.output_name,
            tmp_encoder_out,
            encoder_name,
            encoder_name,
            self.memory_mask_name,
            self.arity,
        )

        return [
            *enc_block_one(),
            *enc_block_two(),
        ]



from neuralogic.core.builder.builder import Builder
from neuralogic.core.builder.dataset_builder import DatasetBuilder
from neuralogic.core.builder.components import Neuron, Atom, Grounding, NeuralSample, NeuronType
from neuralogic.core.builder.dataset import BuiltDataset, GroundedDataset

__all__ = [
    "Builder",
    "DatasetBuilder",
    "Neuron",
    "Atom",
    "Grounding",
    "NeuralSample",
    "NeuronType",
    "BuiltDataset",
    "GroundedDataset",
]



import jpype

from neuralogic.optim.lr_scheduler.lr_decay import LRDecay


class ArithmeticLR(LRDecay):
    r"""
    Decay learning rate on every epoch by the following formula

    .. math::

        \mathbf{lr}_i = \mathbf{lr}_{i-1} - \dfrac{\mathbf{lr}_{0}}{max\_steps}

    Parameters
    ----------

    max_steps : int
    """

    def __init__(self, max_steps: int):
        super().__init__()
        self.max_steps = max_steps

    def _initialize(self, learning_rate):
        class_name = "cz.cvut.fel.ida.neural.networks.computation.training.strategies.Hyperparameters.ArithmeticDecay"
        self._decay = jpype.JClass(class_name)(learning_rate, self.max_steps)

        return self._decay

    def __str__(self) -> str:
        return f"ArithmeticLR(max_steps={self.max_steps})"



from neuralogic.core.constructs.function.function import TransformationFunction, AggregationFunction
from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Transformation, Aggregation, Combination
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class GCNConv(Module):
    r"""
    Graph Convolutional layer from
    `"Semi-supervised Classification with Graph Convolutional Networks" <https://arxiv.org/abs/1609.02907>`_.

    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    edge_name : str
        Edge predicate name to use for neighborhood relations.
    activation : TransformationFunction
        Activation function of the output.
        Default: ``Transformation.IDENTITY``
    aggregation : AggregationFunction
        Aggregation function of nodes' neighbors.
        Default: ``Aggregation.SUM``
    add_self_loops : bool | None
        Add self loops if either set to `True` or `None` (if `normalize` is `True`).
        Default: ``None``
    normalize : bool
        Add normalization.
        Default : ``True``
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        edge_name: str,
        activation: TransformationFunction = Transformation.IDENTITY,
        aggregation: AggregationFunction = Aggregation.SUM,
        add_self_loops: bool | None = None,
        normalize: bool = True,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.edge_name = edge_name

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.aggregation = aggregation
        self.activation = activation

        if add_self_loops is None:
            add_self_loops = normalize

        self.normalize = normalize
        self.add_self_loops = add_self_loops

    def __call__(self):
        head = R.get(self.output_name)(V.I)[self.out_channels, self.in_channels]
        metadata = Metadata(aggregation=self.aggregation, combination=Combination.PRODUCT)

        edge = R.get(self.edge_name)
        edge_count = R.get(f"{self.output_name}__edge_count")

        self_loops = []
        normalization = []
        body = [R.get(self.feature_name)(V.J), edge(V.J, V.I)]

        if self.add_self_loops:
            edge = R.get(f"{self.output_name}__edge")

            self_loops = [
                edge(V.I, V.I)[1.0].fixed(),
                edge(V.I, V.J) <= (R.get(self.edge_name)(V.I, V.J)),
            ]

        if self.normalize:
            count_metadata = Metadata(aggregation=Aggregation.COUNT)
            body = [R.get(self.feature_name)(V.J), edge(V.J, V.I), Transformation.SQRT(edge_count(V.J, V.I))]

            normalization = [
                (edge_count(V.I, V.J) <= edge(V.J, V.X)) | count_metadata,
                (edge_count(V.I, V.J) <= edge(V.I, V.X)) | count_metadata,
                edge_count / 2 | Metadata(combination=Combination.PRODUCT, transformation=Transformation.INVERSE),
            ]

        return [
            *self_loops,
            *normalization,
            (head <= body) | metadata,
            R.get(self.output_name) / 1 | Metadata(transformation=self.activation),
        ]



import io
from typing import List

import pytest

from neuralogic.dataset import CSVFile, CSVDataset, Mode


def test_csv_file_index_in_header() -> None:
    header = ["a", "b", "c"]

    assert CSVFile._find_index_in_header(header, "a") == 0
    assert CSVFile._find_index_in_header(header, "b") == 1
    assert CSVFile._find_index_in_header(header, "c") == 2

    try:
        CSVFile._find_index_in_header(header, "z")
        assert False
    except ValueError:
        assert True


def tests_csv_file() -> None:
    csv_string_source = """a,b,c
    1,2,3
    6,r,s
    7,8,9
    10,11,12
    13,,
    """

    source = io.StringIO(csv_string_source)
    csv_file = CSVFile("my_rel", source)
    examples = csv_file.to_logic_form()

    expected = [
        "my_rel(a, b, c).",
        "my_rel(1, 2, 3).",
        "my_rel(6, r, s).",
        "my_rel(7, 8, 9).",
        "my_rel(10, 11, 12).",
        "my_rel(13, 0, 0).",
    ]

    assert len(examples) == 6
    assert [str(e) for e in examples] == expected

    source = io.StringIO(csv_string_source)
    csv_file = CSVFile("my_rel", source, default_value=1.0, term_columns=[1, 2], skip_rows=1, n_rows=3)
    examples = csv_file.to_logic_form()

    expected = [
        "1.0 my_rel(2, 3).",
        "1.0 my_rel(r, s).",
        "1.0 my_rel(8, 9).",
    ]

    assert len(examples) == 3
    assert [str(e) for e in examples] == expected

    source = io.StringIO(csv_string_source)
    csv_file = CSVFile("my_rel", source, value_column="a", term_columns=["b", "c"], skip_rows=1, header=True, n_rows=2)
    examples = csv_file.to_logic_form()

    expected = [
        "6.0 my_rel(r, s).",
        "7.0 my_rel(8, 9).",
    ]

    assert len(examples) == 2
    assert [str(e) for e in examples] == expected


@pytest.mark.parametrize(
    "mode,expected",
    [
        (
            Mode.ONE_EXAMPLE,
            [
                [
                    "rel_a(1, 2, 3).",
                    "rel_a(4, 5, 6).",
                    "rel_a(7, 8, 9).",
                    "rel_b(1, 2, 3).",
                    "rel_b(4, 5, 6).",
                    "rel_b(7, 8, 9).",
                    "rel_c(1, 2, 3).",
                    "rel_c(4, 5, 6).",
                    "rel_c(7, 8, 9).",
                ]
            ],
        ),
        (
            Mode.ZIP,
            [
                ["rel_a(1, 2, 3).", "rel_b(1, 2, 3).", "rel_c(1, 2, 3)."],
                ["rel_a(4, 5, 6).", "rel_b(4, 5, 6).", "rel_c(4, 5, 6)."],
                ["rel_a(7, 8, 9).", "rel_b(7, 8, 9).", "rel_c(7, 8, 9)."],
            ],
        ),
        (
            Mode.EXAMPLE_PER_SOURCE,
            [
                [
                    "rel_a(1, 2, 3).",
                    "rel_a(4, 5, 6).",
                    "rel_a(7, 8, 9).",
                ],
                [
                    "rel_b(1, 2, 3).",
                    "rel_b(4, 5, 6).",
                    "rel_b(7, 8, 9).",
                ],
                [
                    "rel_c(1, 2, 3).",
                    "rel_c(4, 5, 6).",
                    "rel_c(7, 8, 9).",
                ],
            ],
        ),
    ],
)
def test_csv_dataset(mode: Mode, expected: List[List[str]]) -> None:
    csv_string_source = """1,2,3
    4,5,6
    7,8,9
    """

    source = io.StringIO(csv_string_source)
    csv_source_a = CSVFile("rel_a", source)

    source = io.StringIO(csv_string_source)
    csv_source_b = CSVFile("rel_b", source)

    source = io.StringIO(csv_string_source)
    csv_source_c = CSVFile("rel_c", source)

    dataset = CSVDataset([csv_source_a, csv_source_b, csv_source_c], mode=mode)
    logic_dataset = dataset.to_dataset()

    assert len(logic_dataset) == len(expected)

    for exp, sample in zip(expected, logic_dataset.samples):
        assert len(exp) == len(sample)
        assert exp == [str(e) for e in sample.example]



from typing import Iterable

from neuralogic.core.constructs.function.function import (
    Function,
    AggregationFunction,
    TransformationFunction,
    CombinationFunction,
)


class Metadata:
    __slots__ = "learnable", "transformation", "aggregation", "duplicate_grounding", "combination"

    def __init__(
        self,
        learnable: bool | None = None,
        transformation: TransformationFunction | CombinationFunction | None = None,
        combination: CombinationFunction | None = None,
        aggregation: AggregationFunction | None = None,
        duplicate_grounding: bool | None = None,
    ):
        self.learnable = learnable
        self.combination = combination
        self.transformation = transformation
        self.aggregation = aggregation
        self.duplicate_grounding = duplicate_grounding

    @staticmethod
    def from_iterable(iterable: Iterable) -> "Metadata":
        metadata = Metadata()

        for entry in iterable:
            if callable(entry) and not isinstance(entry, Function):
                entry = entry()
            if isinstance(entry, AggregationFunction):
                metadata.aggregation = entry
            elif isinstance(entry, TransformationFunction):
                metadata.transformation = entry
            elif isinstance(entry, CombinationFunction):
                metadata.combination = entry
            else:
                raise ValueError(f"Invalid entry for metadata: {entry}")
        return metadata

    def __str__(self) -> str:
        metadata_list = []
        if self.learnable is not None:
            metadata_list.append(f"learnable={str(self.learnable).lower()}")
        if self.transformation is not None:
            metadata_list.append(f"transformation={str(self.transformation)}")
        if self.combination is not None:
            metadata_list.append(f"combination={str(self.combination)}")
        if self.aggregation is not None:
            metadata_list.append(f"aggregation={str(self.aggregation)}")
        if self.duplicate_grounding is not None:
            metadata_list.append(f"duplicate_grounding={str(self.duplicate_grounding)}")
        return f"[{', '.join(metadata_list)}]"

    def __repr__(self) -> str:
        return self.__str__()

    def combine(self, other: "Metadata") -> "Metadata":
        return Metadata(
            learnable=other.learnable if other.learnable is not None else self.learnable,
            transformation=other.transformation if other.transformation is not None else self.transformation,
            combination=other.combination if other.combination is not None else self.combination,
            aggregation=other.aggregation if other.aggregation is not None else self.aggregation,
            duplicate_grounding=other.duplicate_grounding
            if other.duplicate_grounding is not None
            else self.duplicate_grounding,
        )

    def __add__(self, other: "Metadata") -> "Metadata":
        return self.combine(other)

    def copy(self) -> "Metadata":
        return Metadata(
            learnable=self.learnable,
            transformation=self.transformation,
            combination=self.combination,
            aggregation=self.aggregation,
            duplicate_grounding=self.duplicate_grounding,
        )



from typing import Dict

import jpype

from neuralogic.core.constructs.function.enum import Combination
from neuralogic.core.constructs.function.function_graph import FunctionGraph
from neuralogic.core.constructs.function.function import Function


class FContainer:
    __slots__ = "nodes", "function"

    def __init__(self, nodes, function: Function):
        self.function = function
        self.nodes = nodes if not self.function.can_flatten else self.get_flattened_nodes(nodes, function)

    @staticmethod
    def get_flattened_nodes(nodes, function: Function):
        new_nodes = []
        for node in nodes:
            if not isinstance(node, FContainer):
                new_nodes.append(node)
                continue

            if node.function.name == function.name:
                new_nodes.extend(node.nodes)
            else:
                new_nodes.append(node)
        return tuple(new_nodes)

    def __add__(self, other):
        return FContainer((self, other), Combination.SUM)

    def __mul__(self, other):
        return FContainer((self, other), Combination.ELPRODUCT)

    def __matmul__(self, other):
        return FContainer((self, other), Combination.PRODUCT)

    def __str__(self):
        if self.function.operator is not None:
            return f" {self.function.operator} ".join(
                node.to_str(True) if isinstance(node, FContainer) else node.to_str() for node in self.nodes
            )

        args = ", ".join(node.to_str() for node in self.nodes)

        if args:
            return f"{self.function}({args})"
        return f"{self.function}"

    @property
    def name(self):
        args = ", ".join(str(node.function) for node in self.nodes if isinstance(node, FContainer))

        if args:
            return f"{self.function}({args})"
        return f"{self.function}"

    def __iter__(self):
        for node in self.nodes:
            if isinstance(node, FContainer):
                for a in node:
                    yield a
            else:
                yield node

    def to_function(self) -> Function:
        graph = self._get_function_node({}, 0)
        return FunctionGraph(name=self.name, function_graph=graph)

    def _get_function_node(self, input_counter: Dict[int, int], start_index: int = 0):
        from neuralogic.core.constructs.relation import BaseRelation

        next_indices = [-1] * len(self.nodes)
        next_nodes = [None] * len(self.nodes)

        for i, node in enumerate(self.nodes):
            if isinstance(node, FContainer):
                next_node = node._get_function_node(input_counter)
                if next_node is None:
                    continue
                next_nodes[i] = next_node
            elif isinstance(node, BaseRelation):
                idx = id(node)

                if node.predicate.hidden or node.predicate.special or node.predicate.name.startswith("_"):
                    continue

                if idx not in input_counter:
                    input_counter[idx] = len(input_counter) + start_index
                next_indices[i] = input_counter[idx]
            else:
                raise ValueError(f"{node} of type {type(node)} inside of body function is not supported")

        filtered_next_node = []
        filtered_next_indices = []

        for i, (node, index) in enumerate(zip(next_nodes, next_indices)):
            if node is not None or index != -1:
                filtered_next_node.append(node)
                filtered_next_indices.append(index)

        if not filtered_next_node or not filtered_next_indices:
            return None

        class_name = "cz.cvut.fel.ida.algebra.functions.combination.FunctionGraph.FunctionGraphNode"

        return jpype.JClass(class_name)(self.function.get(), filtered_next_node, filtered_next_indices)

    def to_str(self, parentheses_wrap: bool = False):
        if parentheses_wrap and self.function.operator is not None:
            return f"({self})"
        return self.__str__()



from typing import Sequence

from neuralogic.core.constructs.metadata import Metadata


class Predicate:
    __slots__ = "name", "arity", "hidden", "special"

    def __init__(self, name, arity, hidden=False, special=False):
        if name.startswith("_"):
            name = name[1:]
            hidden = True

        self.name = name
        self.arity = arity
        self.hidden = hidden
        self.special = special

    def set_arity(self, arity):
        if self.arity == arity:
            return self

    def to_str(self):
        if not self.special and not self.hidden:
            return self.name

        special = "@" if self.special else ""
        hidden = "*" if self.hidden else ""
        return f"{hidden}{special}{self.name}"

    def __str__(self) -> str:
        special = "@" if self.special else ""
        hidden = "*" if self.hidden else ""
        return f"{hidden}{special}{self.name}/{self.arity}"

    def __repr__(self) -> str:
        return self.__str__()

    def __or__(self, other) -> "PredicateMetadata":
        if isinstance(other, Sequence):
            other = Metadata.from_iterable(other)
        elif not isinstance(other, Metadata):
            raise NotImplementedError
        return PredicateMetadata(self, other)


class PredicateMetadata:
    __slots__ = "predicate", "metadata"

    def __init__(self, predicate: Predicate, metadata: Metadata):
        if metadata.aggregation is not None:
            raise ValueError(f"Cannot set 'aggregation' parameter on predicate ({predicate}) metadata")

        if metadata.learnable is not None:
            raise ValueError(f"Cannot set 'learnable' parameter on predicate ({predicate}) metadata")

        self.predicate = predicate
        self.metadata = metadata

    def __str__(self) -> str:
        return f"{self.predicate} {self.metadata}"

    def __repr__(self) -> str:
        return self.__str__()



import jpype

from neuralogic.optim.lr_scheduler.lr_decay import LRDecay


class GeometricLR(LRDecay):
    r"""
    Decay learning rate on every :math:`steps` epoch by the following formula

    .. math::

        \mathbf{lr}_i = \mathbf{lr}_{i-1} \cdot decay\_rate

    Parameters
    ----------

    decay_rate : float

    steps : int

    """

    def __init__(self, decay_rate: float, steps: int):
        super().__init__()
        self.decay_rate = decay_rate
        self.steps = steps

    def _initialize(self, learning_rate):
        class_name = "cz.cvut.fel.ida.neural.networks.computation.training.strategies.Hyperparameters.GeometricDecay"
        self._decay = jpype.JClass(class_name)(learning_rate, self.decay_rate, self.steps)

        return self._decay

    def __str__(self) -> str:
        return f"GeometricLR(decay_rate={self.decay_rate}, steps={self.steps})"
